;
; Montezuma Micro CP/M v 2.32
; BDOS - Basic Disk Operating System
; for Radio Shack Model 4 computer.
;
; This source code is for the Z80ASM assembler
; from SLR systems.
;
	include	system.inc	; get system definitions
;
biosvec	equ	0
iobyte	equ	3
;
defaultdma	equ	80h	; default dma address
;
wboot	equ	biosbase+03h	; BDOS function 0 - system reset
const	equ	biosbase+06h
conin	equ	biosbase+09h
conout	equ	biosbase+0ch
list	equ	biosbase+0fh	; BDOS function 5 - list output
punch	equ	biosbase+12h	; BDOS function 4 - punch output
reader	equ	biosbase+15h
home	equ	biosbase+18h
seldsk	equ	biosbase+1bh
settrk	equ	biosbase+1eh
setsec	equ	biosbase+21h
setdma	equ	biosbase+24h
read	equ	biosbase+27h
write	equ	biosbase+2ah
sectrn	equ	biosbase+30h
;
ctrlc	equ	3
ctrle	equ	5
backsp	equ	8
tab	equ	9
cr	equ	13
lf	equ	10
ctrlp	equ	16
ctrlr	equ	18
ctrls	equ	19
ctrlu	equ	21
ctrlx	equ	24
del	equ	7fh
;
nfuncts	equ	41	; number of bdos functions
;
	org	bdosbase
;
serialnumber:	db	0,0,0,0,0,0
;
	jp	bdosentry
;
; BDOS error table
;
badsctr:	dw	error1	; bad sector
badslct:	dw	error2	; bad disk select
rodisk:		dw	error3	; disk is read only
rofile:		dw	error4	; file is read only
;
; C is function #, parameters in E or DE
; 
bdosentry:
	ex	de,hl
	ld	(params),hl	; save parameters
	ex	de,hl
	ld	a,e
	ld	(eparam),a	; save E parameter separately
	ld	hl,0
	ld	(status),hl	; clear return status
	add	hl,sp
	ld	(usrstack),hl	; save user stack pointer
	ld	sp,bdosstack	; set bdos stack
	xor	a
	ld	(autoflag),a
	ld	(auto),a
	ld	hl,goback	; set return address
	push	hl
	ld	a,c		; get function number
	cp	nfuncts		; valid function?
	ret	nc		; abort if not
	ld	c,e
	ld	hl,functions	; fetch address from function table
	ld	e,a
	ld	d,0
	add	hl,de		; word offset of function address
	add	hl,de
	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ld	hl,(params)	; get the parameters
	ex	de,hl
	jp	(hl)		; and execute the function
;
; BDOS function vector table
;
functions:
	dw	wboot		;  0 system reset
	dw	getcon		;  1 console input
	dw	outcon		;  2 console output
	dw	getrdr		;  3 reader input
	dw	punch		;  4 punch output
	dw	list		;  5 list output
	dw	dircio		;  6 direct console I/O
	dw	getiob		;  7 get I/O byte
	dw	setiob		;  8 set I/O byte
	dw	prtstr		;  9 print string
	dw	rdbuff		; 10 read console buffer
	dw	getcsts		; 11 get console status
	dw	getver		; 12 return version number
	dw	rstdsk		; 13 reset disk system
	dw	setdsk		; 14 select disk
	dw	openfile	; 15 open file
	dw	closefile	; 16 close file
	dw	getfirst	; 17 search for first
	dw	getnxt		; 18 search for next
	dw	delfile		; 19 delete file
	dw	readseq		; 20 read sequential
	dw	writeseq	; 21 write sequential
	dw	fcreate		; 22 make file
	dw	renfile		; 23 rename file
	dw	getlog		; 24 return login vector
	dw	getcrnt		; 25 return current disk
	dw	putdma		; 26 set DMA address
	dw	getaloc		; 27 get addr (alloc)
	dw	wrtprtd		; 28 write protect disk
	dw	getrov		; 29 get R/O vector
	dw	setattr		; 30 set file attributes
	dw	getparm		; 31 get addr (disk parms)
	dw	getuser		; 32 set/get user code
	dw	rdrandom	; 33 read random
	dw	wrrandom	; 34 write random
	dw	filesize	; 35 compute file size
	dw	settran		; 36 set random record
	dw	logoff		; 37 reset drive
	dw	rtn		; 38 reserved
	dw	rtn		; 39 reserved
	dw	wtspecl		; 40 write random with zero fill
;
error1:	ld	hl,badsec	; bad sector on  read or write
	call	prterr
	cp	ctrlc
	jp	z,biosvec	; reboot if ^C
	ret			; else retry read/write
;
error2:	ld	hl,badsel	; bad disk select
	jp	error5		; output message and reboot
;
error3:	ld	hl,diskro	; disk read only error
	jp	error5		; output message and reboot
;
error4:	ld	hl,filero	; file read only error
error5:	call	prterr
	jp	biosvec		; reboot
;
bdoserr:	db	'Bdos Err On '
bdosdrv:	db	' : $'
badsec:	db	'Bad Sector$'
badsel:	db	'Select$'
filero:	db	'File '
diskro:	db	'R/O$'
; 
; Print bdos error message
; HL points to error message
;
prterr:	push	hl
	call	outcrlf		; output new line
	ld	a,(active)
	add	a,'A'
	ld	(bdosdrv),a
	ld	bc,bdoserr
	call	prtmsg		; output bdos error message
	pop	bc
	call	prtmsg		; output passed error message
; 
; Get input character in A
; 
getchar:
	ld	hl,charbuf	; get character from buffer, if any
	ld	a,(hl)
	ld	(hl),0		; clear buffer character
	or	a
	ret	nz		; return char from buffer, if any
	jp	conin		; else return console character
;
; Input and echo a character
; 
getecho:
	call	getchar		; get input character
	call	chkchar		; control character?
	ret	c		; don't echo if so
	push	af		; else save character
	ld	c,a
	call	outcon		; else echo character to console
	pop	af		; and return it
	ret
;
; Check char in A. Set Z if CR, C if other control char
; 
chkchar:
	cp	cr
	ret	z
	cp	lf
	ret	z
	cp	tab
	ret	z
	cp	backsp
	ret	z
	cp	' '
	ret
;
; Check input during output.
; Halt on ^S, reboot on ^C, else put char in charbuf
; 
ckconsol:
	ld	a,(charbuf)	; char in input buffer
	or	a
	jp	nz,ckcon2	; go if char available
	call	const		; else see if console input
	and	1
	ret	z		; done if no console input
	call	conin		; else get the character
	cp	ctrls
	jp	nz,ckcon1	; if not ^S, go return char
	call	conin		; else wait for next console input
	cp	ctrlc
	jp	z,biosvec	; abort if ^C
	xor	a		; ignore input and continue
	ret
;
ckcon1:	ld	(charbuf),a	; put input into buffer
ckcon2:	ld	a,1		; and notify caller
	ret
;
; Output character in C to console
;
outchar:
	ld	a,(outflag)	; check output flag
	or	a
	jp	nz,outchr1
	push	bc
	call	ckconsol	; check console input (for ^S)
	pop	bc
	push	bc
	call	conout		; output C to console
	pop	bc
	push	bc
	ld	a,(prtflag)	; check printer flag
	or	a
	call	nz,list		; and print if requested
	pop	bc
outchr1:
	ld	a,c
	ld	hl,curpos	; get cursor position
	cp	del
	ret	z		; ignore delete character (7fh)
	inc	(hl)		; else increment cursor position
	cp	' '
	ret	nc		; return if not control character
	dec	(hl)		; else adjust cursor position back
	ld	a,(hl)
	or	a
	ret	z		; done if at start of line
	ld	a,c
	cp	backsp
	jp	nz,outchr2	; go if not backspace
	dec	(hl)		; else decrement cursor position
	ret
;
outchr2:
	cp	lf
	ret	nz		; done if not line feed
	ld	(hl),0		; else set cursor to start of line
	ret
;
; Output A to screen
; 
showit:	ld	a,c
	call	chkchar		; is it a control character?
	jp	nc,outcon	; go if not
	push	af
	ld	c,'^'		; else output ^
	call	outchar
	pop	af
	or	'@'		; make control char into letter
	ld	c,a		; and output it
;
; BDOS function 2 - console output
; Output A to screen, expand tabs
;
outcon:	ld	a,c
	cp	tab
	jp	nz,outchar	; go if not a tab
outcon1:
	ld	c,' '		; else expand with spaces
	call	outchar
	ld	a,(curpos)	; out to next tab stop
	and	7
	jp	nz,outcon1
	ret
;
; Output a backspace
; 
backup:	call	backup1		; back up one space
	ld	c,' '		; then blank at cursor position
	call	conout
backup1:
	ld	c,backsp	; and backup one space again
	jp	conout
;
; Signal a deleted line
;
newline:
	ld	c,'#'		; show line deleted
	call	outchar
	call	outcrlf		; output a newline
newln1:	ld	a,(curpos)
	ld	hl,starting	; and blank that line on console
	cp	(hl)
	ret	nc
	ld	c,' '
	call	outchar
	jp	newln1
;
; Output a newline
; 
outcrlf:
	ld	c,cr
	call	outchar
	ld	c,lf
	jp	outchar
;
; Output string at (BC), '$' terminated
; 
prtmsg:	ld	a,(bc)
	cp	'$'
	ret	z		; done if '$' terminator
	inc	bc
	push	bc
	ld	c,a
	call	outcon		; output character from string
	pop	bc
	jp	prtmsg		; loop til done
;
; BDOS function 10 - read console buffer
; 
rdbuff:	ld	a,(curpos)
	ld	(starting),a
	ld	hl,(params)	; get pointer to buffer
	ld	c,(hl)		; first byte is max char count
	inc	hl
	push	hl		; save pointer to return count
	ld	b,0		; init character counter
rdbuf1:	push	bc
	push	hl
rdbuf2:	call	getchar		; get console input
	and	7fh		; clear high bit
	pop	hl
	pop	bc
	cp	cr
	jp	z,rdbuf17	; go if end of input (cr)
	cp	lf
	jp	z,rdbuf17	; go if end of input (lf)
	cp	backsp
	jp	nz,rdbuf3	; go if not backspace
	ld	a,b
	or	a
	jp	z,rdbuf1	; ignore backsp if at start of line
	dec	b		; else decrement character count
	ld	a,(curpos)
	ld	(outflag),a
	jp	rdbuf10
;
rdbuf3:	cp	del
	jp	nz,rdbuf4	; go if not delete key
	ld	a,b
	or	a
	jp	z,rdbuf1	; ignore delete if a start of line
	ld	a,(hl)		; else echo previous character
	dec	b		; and adjust counter and pointer
	dec	hl
	jp	rdbuf15
;
rdbuf4:	cp	ctrle
	jp	nz,rdbuf5	; go if not end of line
	push	bc
	push	hl
	call	outcrlf		; else output newline
	xor	a
	ld	(starting),a	; start over
	jp	rdbuf2
;
rdbuf5:	cp	ctrlp
	jp	nz,rdbuf6	; go if not print toggle
	push	hl
	ld	hl,prtflag
	ld	a,1
	sub	(hl)
	ld	(hl),a		; toggle print flag
	pop	hl
	jp	rdbuf1
;
rdbuf6:	cp	ctrlx
	jp	nz,rdbuf8	; go if not cancel key
	pop	hl
rdbuf7:	ld	a,(starting)	; else remove current line
	ld	hl,curpos
	cp	(hl)
	jp	nc,rdbuff
	dec	(hl)
	call	backup		; output a backspace
	jp	rdbuf7		; do til start of line
;
rdbuf8:	cp	ctrlu
	jp	nz,rdbuf9	; go if not remove line key
	call	newline		; else output a newline
	pop	hl
	jp	rdbuff		; and start over
;
rdbuf9:	cp	ctrlr
	jp	nz,rdbuf14	; go if not retype line key
rdbuf10:
	push	bc
	call	newline		; else output a newline
	pop	bc
	pop	hl
	push	hl
	push	bc
rdbuf11:			; output current line
	ld	a,b
	or	a
	jp	z,rdbuf12	; go if end of line
	inc	hl
	ld	c,(hl)		; else output next character
	dec	b
	push	bc
	push	hl
	call	showit
	pop	hl
	pop	bc
	jp	rdbuf11
;
rdbuf12:
	push	hl		; done with line
	ld	a,(outflag)
	or	a
	jp	z,rdbuf2
	ld	hl,curpos	; update cursor position
	sub	(hl)
	ld	(outflag),a
rdbuf13:
	call	backup		; output a backspace
	ld	hl,outflag
	dec	(hl)
	jp	nz,rdbuf13
	jp	rdbuf2		; get the next character
;
rdbuf14:
	inc	hl
	ld	(hl),a		; put character in buffer
	inc	b		; increment character counter
rdbuf15:
	push	bc
	push	hl
	ld	c,a
	call	showit		; echo the character
	pop	hl
	pop	bc
	ld	a,(hl)		; recover character
	cp	ctrlc
	ld	a,b
	jp	nz,rdbuf16	; go if not abort (^C)
	cp	1
	jp	z,biosvec	; reboot only if ^C is first character
rdbuf16:
	cp	c		; test if buffer full
	jp	c,rdbuf1	; go if not
rdbuf17:
	pop	hl
	ld	(hl),b		; store return character count
	ld	c,cr		; and output cr
	jp	outchar
;
; BDOS function 1 - console input
; 
getcon:	call	getecho		; get input character and echo it
	jp	setstat		; save status and return
;
; BDOS function 3 - reader input
; 
getrdr:	call	reader		; get reader character
	jp	setstat		; save status and return
;
; BDOS function 6 - direct console I/O
; 
dircio:	ld	a,c		; test flag for 0ffh
	inc	a
	jp	z,dirc1		; go if input function
	inc	a		; test flag for 0feh
	jp	z,const		; go if status function
	jp	conout		; else is output function
;
dirc1:	call	const		; get console status
	or	a
	jp	z,goback1	; done if no character available
	call	conin		; else get the character
	jp	setstat		; and set return status
;
; BDOS function 7 - get I/O byte
; 
getiob:	ld	a,(iobyte)
	jp	setstat
;
; BDOS function 8 - set I/O byte
; 
setiob:	ld	hl,iobyte
	ld	(hl),c
	ret
;
; BDOS function 9 - print string
; Output string at (DE), terminated by '$'
; 
prtstr:	ex	de,hl
	ld	c,l
	ld	b,h		; BC points to message
	jp	prtmsg		; go print it
;
; BDOS function 11 - get console status
; 
getcsts:
	call	ckconsol
;
; Set function status and return to caller
;
setstat:
	ld	(status),a
rtn:	ret
;
ioerr1:	ld	a,1		; set read/write error code
	jp	setstat
;
outflag:	db	0	; set nonzero to supress output
starting:	db	2	; start position for cursor
curpos:		db	0	; current cursor position
prtflag:	db	0	; printer toggle flag (^P), list if nonzero
;
charbuf:	db	0	; single character input buffer
;
usrstack:	dw	0	; place to save user stack pointer
;
	dw	0,0,0,0,0,0,0,0	; 48 bytes for BDOS stack
	dw	0,0,0,0,0,0,0,0
	dw	0,0,0,0,0,0,0,0
;
bdosstack	equ	$
;
userno:	db	0	; current user number
active:	db	0	; current active disk drive
params:	dw	0	; save DE parameters here
;
status:	dw	0	; BDOS return status
;
; Select error
;
slcterr:
	ld	hl,badslct	; jump to error routine
;
; Indirect jump to (HL)
;
jumphl:	ld	e,(hl)
	inc	hl
	ld	d,(hl)
	ex	de,hl
	jp	(hl)
;
; Block move from (DE) to (HL), for C bytes
; 
de2hl:	inc	c
de2hl1:	dec	c
	ret	z
	ld	a,(de)
	ld	(hl),a
	inc	de
	inc	hl
	jp	de2hl1
;
; Select the drive in ACTIVE
; 
select:	ld	a,(active)
	ld	c,a
	call	seldsk
	ld	a,h
	or	l
	ret	z		; return if invalid drive
; 
; HL -> DPB
; Extract pointers from DPB
; 
	ld	e,(hl)
	inc	hl
	ld	d,(hl)		; DE -> translation table
	inc	hl
	ld	(scratch1),hl	; save pointers in scratch area
	inc	hl
	inc	hl
	ld	(scratch2),hl
	inc	hl
	inc	hl
	ld	(scratch3),hl
	inc	hl
	inc	hl
	ex	de,hl
	ld	(xlate),hl	; save translation table address
	ld	hl,dirbuf	; transfer dpb, parameter block pointer,
	ld	c,8		; check vector and alocation vectors
	call	de2hl
	ld	hl,(diskpb)
	ex	de,hl
	ld	hl,sectors	; copy parameter block
	ld	c,15
	call	de2hl
	ld	hl,(dsksize)
	ld	a,h
	ld	hl,bigdisk
	ld	(hl),0ffh	; set small disk
	or	a		; more than 256 blocks on disk?
	jp	z,select1	; go if not
	ld	(hl),0		; else set big disk
select1:
	ld	a,0ffh
	or	a
	ret
;
; Home disk track and clear pointers
; 
homedrv:
	call	home		; disk head to track 0
	xor	a
	ld	hl,(scratch2)	; zero track pointer
	ld	(hl),a
	inc	hl
	ld	(hl),a
	ld	hl,(scratch3)	; zero sector pointer
	ld	(hl),a
	inc	hl
	ld	(hl),a
	ret
;
; Do disk read and check return status
; 
doread:	call	read
	jp	ioret
;
; Do disk write and handle any bios error
; 
dowrite:
	call	write
ioret:	or	a
	ret	z		; return if no error
	ld	hl,badsctr
	jp	jumphl
;
; Select track & sector block number falls in
; 
trksec:	ld	hl,(filepos)	; get position of last accessed file
	ld	c,2		; in the directory, and compute sector
	call	shiftr		; sector = file - position / 4
	ld	(blknum),hl	; save block number
	ld	(cksumtbl),hl	; don't know why it goes in chksum table
trksec1:
	ld	hl,blknum
	ld	c,(hl)		; get sector number into BC
	inc	hl
	ld	b,(hl)
	ld	hl,(scratch3)
	ld	e,(hl)		; get current sector into DE
	inc	hl
	ld	d,(hl)
	ld	hl,(scratch2)
	ld	a,(hl)		; get current track number into HL
	inc	hl
	ld	h,(hl)
	ld	l,a
trksec2:
	ld	a,c
	sub	e
	ld	a,b		; is sector before current one?
	sbc	a,d
	jp	nc,trksec3	; go if not
	push	hl		; else decrement sectors by one track
	ld	hl,(sectors)	; get sectors per track
	ld	a,e
	sub	l
	ld	e,a
	ld	a,d
	sbc	a,h
	ld	d,a		; backed up one track
	pop	hl
	dec	hl		; adjust track counter
	jp	trksec2
;
trksec3:
	push	hl		; sector is after current sector
	ld	hl,(sectors)	; get sectors per track
	add	hl,de		; increase sector pointer to next track
	jp	c,trksec4	; go if sector in this track
	ld	a,c
	sub	l
	ld	a,b
	sbc	a,h
	jp	c,trksec4	; go if sector in this track
	ex	de,hl		; if not, increment track counter
	pop	hl
	inc	hl
	jp	trksec3		; continue until correct track
;
trksec4:
	pop	hl		; track number in HL
	push	bc
	push	de
	push	hl
	ex	de,hl
	ld	hl,(offset)	; adjust for track offset
	add	hl,de
	ld	b,h
	ld	c,l
	call	settrk		; select this track
	pop	de
	ld	hl,(scratch2)	; set track pointer
	ld	(hl),e
	inc	hl
	ld	(hl),d
	pop	de
	ld	hl,(scratch3)	; set sector pointer
	ld	(hl),e
	inc	hl
	ld	(hl),d
	pop	bc
	ld	a,c
	sub	e
	ld	c,a		; get relative sector into BC
	ld	a,b
	sbc	a,d
	ld	b,a
	ld	hl,(xlate)	; get sector translate table
	ex	de,hl
	call	sectrn		; call bios to translate sector
	ld	c,l
	ld	b,h
	jp	setsec		; select translated sector
;
; Compute block number from record number
; (savnrec) and extent number (savext)
;
getblock:
	ld	hl,blkshft	; get logical to physical shift
	ld	c,(hl)
	ld	a,(savnrec)	; get record number
getblk1:
	or	a		; A = A / (2^blkshft)
	rra
	dec	c
	jp	nz,getblk1
	ld	b,a		; save result
	ld	a,8
	sub	(hl)
	ld	c,a		; C = 8 - blkshft
	ld	a,(savext)
getblk2:
	dec	c		; A = savext * 2^blkshft
	jp	z,getblk3
	or	a
	rla
	jp	getblk2
;
getblk3:
	add	a,b		; return block number
	ret
;
; Extract block byte from the fcb
; BC = offset into block table of fcb
; If big disk, block numbers are 2 bytes,
; else they are one byte.
;
extblk:	ld	hl,(params)	; get fcb address
	ld	de,16
	add	hl,de		; point to block numbers in fcb
	add	hl,bc		; adjust pointer for correct block
	ld	a,(bigdisk)
	or	a
	jp	z,extblk1	; go if big disk
	ld	l,(hl)		; else get 1 byte block number
	ld	h,0
	ret
;
extblk1:
	add	hl,bc		; double offset
	ld	e,(hl)		; and get 2 byte block number
	inc	hl
	ld	d,(hl)
	ex	de,hl
	ret
;
; Compute block number
;
comblk:	call	getblock	; get block number from record number
	ld	c,a
	ld	b,0
	call	extblk		; extract block number from fcb
	ld	(blknum),hl	; and save it
	ret
;
chkblk:	ld	hl,(blknum)	; check for zero block number
	ld	a,l
	or	h
	ret
;
; Convert physical block number (blknum) to logical sector
; number (logsect). Add actual sector to logical sector and
; store result in blknum.
;
logical:
	ld	a,(blkshft)	; get shift factor
	ld	hl,(blknum)	; get physical sector number
logicl1:
	add	hl,hl		; compute logical sector number
	dec	a
	jp	nz,logicl1
	ld	(logsect),hl	; save it
	ld	a,(blkmask)	; get block mask
	ld	c,a
	ld	a,(savnrec)	; get next sector
	and	c		; A = relative position in block
	or	l		; add to logical sector
	ld	l,a
	ld	(blknum),hl	; and save result
	ret
;
; Set HL to point to extent byte in fcb
;
setext:	ld	hl,(params)
	ld	de,12		; offset to extent byte
	add	hl,de
	ret
;
; Set HL to point to record count in fcb and DE to
; point to record number in fcb
;
sethlde:
	ld	hl,(params)
	ld	de,15		; offset to record count
	add	hl,de
	ex	de,hl
	ld	hl,17		; offset to record number
	add	hl,de
	ret
;
; Save record count, record number, and extent from fcb
;
strdata:
	call	sethlde		; set pointers to fcb
	ld	a,(hl)		; get record count from fcb
	ld	(savnrec),a	; and save it
	ex	de,hl
	ld	a,(hl)		; get record number from fcb
	ld	(savnxt),a	; and save it
	call	setext
	ld	a,(extmask)
	and	(hl)		; mask record number
	ld	(savext),a	; save extent
	ret
;
; Set next record to access.
; Mode = 1 for sequential access.
;
setnrec:
	call	sethlde		; get fcb pointers
	ld	a,(mode)
	cp	2
	jp	nz,stnrec1
	xor	a
stnrec1:
	ld	c,a
	ld	a,(savnrec)	; get last record count
	add	a,c		; increment record count
	ld	(hl),a		; and set in fcb record number
	ex	de,hl
	ld	a,(savnxt)	; get next record
	ld	(hl),a		; and put in fcb record count
	ret
;
; Shift HL right by C bits
;
shiftr:	inc	c
shiftr1:
	dec	c
	ret	z
	ld	a,h
	or	a
	rra
	ld	h,a
	ld	a,l
	rra
	ld	l,a
	jp	shiftr1
;
; Compute check sum for buffer
;
checksum:
	ld	c,128		; record length
	ld	hl,(dirbuf)	; point to directory buffer
	xor	a
chksum1:
	add	a,(hl)		; compute check sum
	inc	hl
	dec	c
	jp	nz,chksum1
	ret
;
; Shift HL left by C bits
;
shiftl:	inc	c
shiftl1:
	dec	c
	ret	z
	add	hl,hl
	jp	shiftl1
;
; Set bit in BC corresponding to current drive.
;
setbit:	push	bc
	ld	a,(active)	; get active drive
	ld	c,a
	ld	hl,1
	call	shiftl		; shift bit in HL to correct position
	pop	bc		; or bit with BC and put result in HL
	ld	a,c
	or	l
	ld	l,a
	ld	a,b
	or	h
	ld	h,a
	ret
;
; Get write protect status of current drive.
;
getwprt:
	ld	hl,(wrtprt)	; get write protect bits
	ld	a,(active)	; get current drive number
	ld	c,a
	call	shiftr		; shift drive bit to bit 0
	ld	a,l
	and	1		; mask off wp bit for drive
	ret
;
; BDOS function 28 - write protect disk
;
wrtprtd:
	ld	hl,wrtprt
	ld	c,(hl)		; set BC to write protect bits
	inc	hl
	ld	b,(hl)
	call	setbit		; and set wp bit for drive
	ld	(wrtprt),hl	; save result
	ld	hl,(dirsize)	; get directory size
	inc	hl
	ex	de,hl
	ld	hl,(scratch1)	; and save it
	ld	(hl),e
	inc	hl
	ld	(hl),d
	ret
;
; Check if file is read only
;
chkrofl:
	call	fcb2hl		; HL points to file entry in directory buffer
ckrof1:	ld	de,9		; file R/O is bit 7 of 9th byte in fcb
	add	hl,de
	ld	a,(hl)
	rla
	ret	nc		; return if file is read/write
	ld	hl,rofile	; else print read only message
	jp	jumphl		; and return
;
; Check if disk is read only
;
chkwprt:
	call	getwprt		; get write protect status
	ret	z		; done if disk is read/write
	ld	hl,rodisk	; else print read only message
	jp	jumphl		; and return
;
; Set HL to point to proper entry in directory buffer
;
fcb2hl:	ld	hl,(dirbuf)	; get dir buffer address
	ld	a,(fcbpos)	; get position of file
adda2hl:			; add position to pointer
	add	a,l
	ld	l,a
	ret	nc
	inc	h
	ret
;
; Get the S2 byte from the fcb
;
gets2:	ld	hl,(params)	; get address of fcb
	ld	de,14		; offset to s2 byte
	add	hl,de
	ld	a,(hl)		; get s2 byte from fcb
	ret
;
; Clear the S2 byte in the fcb
;
clears2:
	call	gets2		; get address of s2 byte
	ld	(hl),0		; zero it
	ret
;
; Set bit 7 of S2 byte in the fcb
;
sets2b7:
	call	gets2		; get address of s2 byte
	or	80h
	ld	(hl),a
	ret
;
; Compare file position with scratch data and set flags.
; Checks if there are more file names in the directory.
;
morefls:
	ld	hl,(filepos)	; current position
	ex	de,hl
	ld	hl,(scratch1)	; limit
	ld	a,e		; compute difference
	sub	(hl)
	inc	hl
	ld	a,d
	sbc	a,(hl)		; set carry if no more files
	ret
;
chknmbr:
	call	morefls		; DE = filepos, HL -> scratch1
	ret	c		; done if no more files
	inc	de
	ld	(hl),d		; else update scratch1
	dec	hl
	ld	(hl),e
	ret
;
; HL = DE - HL
;
subhl:	ld	a,e
	sub	l
	ld	l,a
	ld	a,d
	sbc	a,h
	ld	h,a
	ret
;
; Set directory checksum
;
setdir:	ld	c,0ffh
;
; Set or compare directory checksum. If C = 0ffh on entry,
; set the checksum, else check it. If the check fails,
; indicating the disk has been changed, then write protect
; the disk.
;
checkdir:
	ld	hl,(cksumtbl)
	ex	de,hl
	ld	hl,(alloc1)
	call	subhl
	ret	nc		; ok if chksumtbl > alloc1
	push	bc
	call	checksum	; else compute checksum
	ld	hl,(chkvect)
	ex	de,hl
	ld	hl,(cksumtbl)
	add	hl,de		; HL -> byte for this drive
	pop	bc
	inc	c
	jp	z,chkdir1	; go if set function
	cp	(hl)		; else check it
	ret	z		; ok if they are the same
	call	morefls		; are there more files?
	ret	nc		; done if no more files
	call	wrtprtd		; else write protect the disk
	ret
;
chkdir1:
	ld	(hl),a		; set write protect for this disk
	ret
;
; Write to directory of current disk
;
dirwrite:
	call	setdir		; set checksum
	call	dirdma		; set DMA address
	ld	c,1		; force write
	call	dowrite		; BIOS write
	jp	defdma
;
; Read from directory
;
dirread:
	call	dirdma		; set directory DMA address
	call	doread		; read it
defdma:	ld	hl,userdma	; reset DMA to user's address
	jp	dirdma1
;
; Set DMA address for directory
;
dirdma:	ld	hl,dirbuf
dirdma1:
	ld	c,(hl)		; get DMA address into BC
	inc	hl
	ld	b,(hl)
	jp	setdma		; BIOS set DMA
;
; Move directory buffer into user DMA space
;
movedir:
	ld	hl,(dirbuf)	; get directory buffer address
	ex	de,hl		; into DE
	ld	hl,(userdma)	; user's DMA address into HL
	ld	c,128		; dir buffer length
	jp	de2hl		; transfer the data
;
; Check filepos and set Z if filepos = 0ffffh
;
ckfilpos:
	ld	hl,filepos
	ld	a,(hl)
	inc	hl
	cp	(hl)
	ret	nz
	inc	a
	ret
;
; Set filepos to 0ffffh
;
stfilpos:
	ld	hl,-1
	ld	(filepos),hl
	ret
;
; Move to next file position in directory buffer. If no more,
; set pointer to -1.
; On entry: C = 0ffh to set checksum, else check the disk and
; set write protect if checksums are not equal.
;
nxentry:
	ld	hl,(dirsize)	; get directory size limit
	ex	de,hl
	ld	hl,(filepos)	; get current count
	inc	hl
	ld	(filepos),hl	; set next one
	call	subhl		; HL = dirsize - filepos
	jp	nc,nxent1	; go if there is room left
	jp	stfilpos	; else set flag and return
;
nxent1:	ld	a,(filepos)	; get file position in directory
	and	3		; 0-4 entries only
	ld	b,5		; get relative position
nxent2:	add	a,a		; A = A shift left by five bits
	dec	b
	jp	nz,nxent2
	ld	(fcbpos),a	; save position in fcb
	or	a
	ret	nz		; done if within buffer
	push	bc
	call	trksec		; else get next directory entry
	call	dirread
	pop	bc
	jp	checkdir
;
; Get a bit from disk space allocation
; Return result in A.
; On entry: D = bit position, HL -> address
;           BC = block number to check
;
ckbitmap:
	ld	a,c		; get bit number
	and	7		; compute D = E = (C & 7) + 1
	inc	a
	ld	e,a		; save bit number
	ld	d,a
	ld	a,c		; BC /= 8
	rrca			; shift right 3 bits
	rrca
	rrca
	and	1fh		; clear bits 5-7
	ld	c,a
	ld	a,b
	add	a,a		; shift B into bits 5-7
	add	a,a
	add	a,a
	add	a,a
	add	a,a
	or	c		; add in C
	ld	c,a
	ld	a,b
	rrca
	rrca
	rrca
	and	1fh
	ld	b,a		; BC is offset into allocation table
	ld	hl,(allocvect)
	add	hl,bc
	ld	a,(hl)		; get correct byte
ckbmap1:
	rlca			; put bit in position 0
	dec	e
	jp	nz,ckbmap1
	ret
;
; Set or clear bit map so that block number in BC will be
; marked used.
; On entry E = 0 to clear bit, 1 to set bit.
;
stbitmap:
	push	de
	call	ckbitmap	; get the correct byte
	and	0feh		; clear bit
	pop	bc
	or	c		; set according to C
;
; Restore bit position and then store in table. A = value,
; D = bit position (1-8), HL -> address within allocation table.
;
stbmap1:
	rrca			; restore bit position
	dec	d
	jp	nz,stbmap1
	ld	(hl),a		; store byte in alloc table
	ret
;
; Set or clear space used bits in allocation map for file.
; On entry C = 0 to clear it, 1 to set it.
;
setfile:
	call	fcb2hl		; get address of fcb
	ld	de,16
	add	hl,de		; point to block number
	push	bc
	ld	c,17		; check all bytes in table
setfl1:	pop	de
	dec	c
	ret	z		; return if done
	push	de
	ld	a,(bigdisk)	; check for 16 bit block numbers
	or	a
	jp	z,setfl2	; go if 16 bit blk num
	push	bc
	push	hl
	ld	c,(hl)		; else get 8 bit block number
	ld	b,0
	jp	setlf3
;
setfl2:	dec	c		; adjust counter for 16 bit blknum
	push	bc
	ld	c,(hl)		; get low byte
	inc	hl
	ld	b,(hl)		; get high byte
	push	hl
setlf3:	ld	a,c		; is this block used?
	or	b
	jp	z,setfl4	; go if not
	ld	hl,(dsksize)	; is block within disk space?
	ld	a,l
	sub	c
	ld	a,h
	sbc	a,b
	call	nc,stbitmap	; set proper bit
setfl4:	pop	hl		; point to next block num in fcb
	inc	hl
	pop	bc
	jp	setfl1
;
; Construct space used allocation bit map for active drive.
;
bitmap:	ld	hl,(dsksize)	; compute size of alloc table
	ld	c,3
	call	shiftr		; divide HL by 8
	inc	hl
	ld	b,h
	ld	c,l		; set BC to alloc table length
	ld	hl,(allocvect)
;
; Initialize bitmap for this drive.
;
bitmap1:
	ld	(hl),0		; zero the table
	inc	hl
	dec	bc
	ld	a,b
	or	c
	jp	nz,bitmap1
	ld	hl,(alloc0)	; get initial space used by directory
	ex	de,hl
	ld	hl,(allocvect)	; and put into map
	ld	(hl),e
	inc	hl
	ld	(hl),d
	call	homedrv		; home the drive
	ld	hl,(scratch1)
	ld	(hl),3		; force next directory request to read a sector
	inc	hl
	ld	(hl),0
	call	stfilpos	; clear inital file position
bitmap2:
	ld	c,0ffh
	call	nxentry		; read next file name in directory
	call	ckfilpos	; is there another file?
	ret	z		; done if not
	call	fcb2hl		; else get its address
	ld	a,0e5h
	cp	(hl)		; empty file
	jp	z,bitmap2	; go if so
	ld	a,(userno)	; else check user number
	cp	(hl)
	jp	nz,bitmap3	; go if not correct user
	inc	hl
	ld	a,(hl)
	sub	'$'
	jp	nz,bitmap3	; go if name doesn't start with '$'
	dec	a
	ld	(status),a	; else set its status to -1
bitmap3:
	ld	c,1
	call	setfile		; mark file's space as used in bit map
	call	chknmbr
	jp	bitmap2
;
ststatus:
	ld	a,(fndstat)	; set status and return
	jp	setstat
;
; Check extents in A and C. Set Z flag if they're the same.
;
samext:	push	bc
	push	af
	ld	a,(extmask)	; get extent mask
	cpl
	ld	b,a
	ld	a,c		; mask first extent
	and	b
	ld	c,a		; put back in C
	pop	af
	and	b		; mask second extent
	sub	c		; and compare with first
	and	1fh		; only check bits 0-4
	pop	bc		; Z set if they're the same
	ret
;
; Search for first occurance of file name.
; On entry C = number of bytes in fcb that must match.
;
findfst:
	ld	a,0ffh
	ld	(fndstat),a
	ld	hl,counter
	ld	(hl),c		; save character count
	ld	hl,(params)	; get pointer to file name
	ld	(savefcb),hl	; save pointer
	call	stfilpos	; clear initial file position
	call	homedrv
findnxt:
	ld	c,0		; write protect disk if changed
	call	nxentry		; get next filename in directory
	call	ckfilpos	; is file position -1?
	jp	z,fndnxt6	; done if so
	ld	hl,(savefcb)	; else get pointer to filename to match
	ex	de,hl
	ld	a,(de)
	cp	0e5h
	jp	z,fndnxt1	; go if empty entry
	push	de
	call	morefls		; more files in directory?
	pop	de
	jp	nc,fndnxt6	; done if not
fndnxt1:
	call	fcb2hl		; get address of this fcb in directory
	ld	a,(counter)	; get number of characters to check
	ld	c,a
	ld	b,0
fndnxt2:
	ld	a,c
	or	a
	jp	z,fndnxt5	; go if all bytes checked
	ld	a,(de)
	cp	'?'
	jp	z,fndnxt4	; go if wild card
	ld	a,b		; get byte position in fcb
	cp	13
	jp	z,fndnxt4	; ignore if s1 byte
	cp	12
	ld	a,(de)
	jp	z,fndnxt3	; ignore if extent byte
	sub	(hl)
	and	7fh
	jp	nz,findnxt	; not the same
	jp	fndnxt4		; check next entry
;
fndnxt3:
	push	bc
	ld	c,(hl)		; check the extent byte
	call	samext
	pop	bc
	jp	nz,findnxt	; go if not the same extent
fndnxt4:
	inc	de		; increment byte pointers
	inc	hl
	inc	b
	dec	c		; count down characters
	jp	fndnxt2
;
fndnxt5:
	ld	a,(filepos)	; return position of this entry
	and	3
	ld	(status),a
	ld	hl,fndstat
	ld	a,(hl)
	rla
	ret	nc
	xor	a
	ld	(hl),a
	ret
;
fndnxt6:
	call	stfilpos	; filename not found
	ld	a,0ffh		; set filepos to -1
	jp	setstat		; and status to not found
;
; Erase file from directory. Sets first byte to 0e5h.
;
erafile:
	call	chkwprt		; is disk write protected?
	ld	c,12
	call	findfst		; get first file name
erafil1:
	call	ckfilpos	; file found?
	ret	z		; done if not
	call	chkrofl		; is it read only?
	call	fcb2hl		; if not, get address of fcb
	ld	(hl),0e5h	; and 'erase' file
	ld	c,0		; clear its space from the bit map
	call	setfile
	call	dirwrite	; write the directory sector
	call	findnxt		; find the next file name
	jp	erafil1		; and repeat
;
; Find next available block in allocation map.
;
fndspace:
	ld	d,b		; DE = block to check
	ld	e,c
fndspa1:
	ld	a,c
	or	b
	jp	z,fndspa2	; go if block 0
	dec	bc		; else check previous block
	push	de
	push	bc
	call	ckbitmap	; check if block is empty
	rra
	jp	nc,fndspa3	; block is empty, use it
	pop	bc		; else check some more
	pop	de
fndspa2:
	ld	hl,(dsksize)	; check if block DE is within disk limit
	ld	a,e
	sub	l
	ld	a,d
	sbc	a,h
	jp	nc,fndspa4	; if not, then no free block was found
	inc	de		; else check next one
	push	bc
	push	de
	ld	b,d
	ld	c,e
	call	ckbitmap	; check if block is empty
	rra
	jp	nc,fndspa3	; keep searching if not empty
	pop	de
	pop	bc
	jp	fndspa1
;
fndspa3:			; empty block found
	rla
	inc	a
	call	stbmap1		; update bit map
	pop	hl
	pop	de
	ret
;
fndspa4:
	ld	a,c		; empty block not found
	or	b
	jp	nz,fndspa1	; go if more blocks to test
	ld	hl,0		; else set not found status
	ret
;
; Move an fcb into the directory and write to disk
;
fcbset:	ld	c,0
	ld	e,32		; length of entry
;
update:	push	de
	ld	b,0		; BC = relative byte position
	ld	hl,(params)	; get address of fcb
	add	hl,bc		; point to starting byte
	ex	de,hl
	call	fcb2hl		; get address of fcb in directory
	pop	bc
	call	de2hl		; move fcb to directory
update1:
	call	trksec		; get directory track and sector
	jp	dirwrite	; then go write it
;
; Change name of files on disk. FCB has current name in the
; first 12 characters and the new name at fcb offset 16.
;
chgnames:
	call	chkwprt		; check for write protected disk
	ld	c,12
	call	findfst		; find first name
	ld	hl,(params)	; hl points to fcb
	ld	a,(hl)		; get user number
	ld	de,16
	add	hl,de		; point to desired name
	ld	(hl),a		; set user number to same as old name
chgnam1:
	call	ckfilpos	; matching file found?
	ret	z		; done if not
	call	chkrofl		; check for read only file
	ld	c,16
	ld	e,12		; update first 12 bytes in directory
	call	update
	call	findnxt		; get next file name
	jp	chgnam1		; and loop til done
;
; Update file attributes
;
saveattr:
	ld	c,12		; match first 12 bytes in filename
	call	findfst		; search for first one
savatr1:
	call	ckfilpos	; was it found?
	ret	z		; done if not
	ld	c,0
	ld	e,12
	call	update		; update filename and write directory
	call	findnxt		; search for next match
	jp	savatr1		; and loop til done
;
; Open the file specified in fcb
;
openit:	ld	c,15		; check first 15 bytes
	call	findfst		; search for first
	call	ckfilpos	; any found?
	ret	z		; done if not
openit1:
	call	setext		; else get extent byte
	ld	a,(hl)
	push	af		; save extent byte
	push	hl		; and its address
	call	fcb2hl		; HL points to fcb in directory
	ex	de,hl
	ld	hl,(params)	; get user's fcb
	ld	c,32
	push	de
	call	de2hl		; move fcb to user's space
	call	sets2b7		; set bit 7 in s2
	pop	de		; restore fcb address
	ld	hl,12		; point to extent byte
	add	hl,de
	ld	c,(hl)		; get extent byte
	ld	hl,15
	add	hl,de		; point to record count
	ld	b,(hl)		; and put it in B
	pop	hl
	pop	af
	ld	(hl),a		; set original extent
	ld	a,c
	cp	(hl)		; same extent as in directory fcb?
	ld	a,b
	jp	z,openit2	; if so, use same record count
	ld	a,0		; if user specified extent greater than that
	jp	c,openit2	; in the directory, set record count to 0
	ld	a,128		; else set to maximum
openit2:
	ld	hl,(params)	; HL points to user's fcb
	ld	de,15
	add	hl,de
	ld	(hl),a		; set record count
	ret
;
; Move two bytes from (DE) to (HL), but only if (HL) is 0.
; On return, Z set if word was moved.
;
moveword:
	ld	a,(hl)		; check for both bytes zero
	inc	hl
	or	(hl)
	dec	hl
	ret	nz		; done if both not zero
	ld	a,(de)		; else copy bytes
	ld	(hl),a
	inc	de
	inc	hl
	ld	a,(de)
	ld	(hl),a
	dec	de		; restore register values
	dec	hl
	ret
;
; Close file in fcb
;
closeit:
	xor	a
	ld	(status),a	; clear status and file position
	ld	(filepos),a
	ld	(filepos+1),a
	call	getwprt
	ret	nz		; done if drive is write protected
	call	gets2		; else get s2 byte
	and	80h		; check modified bit
	ret	nz		; done if file not modified
	ld	c,15
	call	findfst		; else find file in directory
	call	ckfilpos
	ret	z		; done if not found
	ld	bc,16
	call	fcb2hl		; HL points to records used
	add	hl,bc
	ex	de,hl
	ld	hl,(params)	; HL points to user's fcb
	add	hl,bc
	ld	c,16
closeit1:
	ld	a,(bigdisk)
	or	a
	jp	z,closeit4	; go if 16 bit block numbers
	ld	a,(hl)		; else get 8 bit blknum from user's fcb
	or	a
	ld	a,(de)
	jp	nz,closeit2	; go if user's blknum not zero
	ld	(hl),a		; else update with blknum from directory
closeit2:
	or	a
	jp	nz,closeit3	; go if directory byte not zero
	ld	a,(hl)		; else update from user's fcb
	ld	(de),a
closeit3:
	cp	(hl)
	jp	nz,closeit7	; close error if bytes not equal
	jp	closeit5
;
closeit4:			; 16 bit block numbers (bigdisk)
	call	moveword	; update user's fcb if it's zero
	ex	de,hl
	call	moveword	; update directory fcb if it's zero
	ex	de,hl
	ld	a,(de)
	cp	(hl)
	jp	nz,closeit7	; close error if words not equal
	inc	de
	inc	hl
	ld	a,(de)		; test second byte
	cp	(hl)
	jp	nz,closeit7
	dec	c
closeit5:
	inc	de		; point to next item in table
	inc	hl
	dec	c
	jp	nz,closeit1	; loop if more to do
	ld	bc,-20		; else back up to extent byte
	add	hl,bc
	ex	de,hl
	add	hl,bc
	ld	a,(de)
	cp	(hl)		; directory extent greater than user's?
	jp	c,closeit6	; go if so
	ld	(hl),a		; else update directory extent
	ld	bc,iobyte	; and update record count in directory fcb
	add	hl,bc
	ex	de,hl
	add	hl,bc
	ld	a,(hl)		; get user's record count byte
	ld	(de),a		; put in directory fcb
closeit6:
	ld	a,0ffh		; set close flag to was open and
	ld	(closeflg),a	; is now closed
	jp	update1		; go update directory
;
closeit7:
	ld	hl,status	; set status and return
	dec	(hl)
	ret
;
;  Get next empty space in directory
;
getempty:
	call	chkwprt		; check if disk is write protected
	ld	hl,(params)	; get current fcb
	push	hl
	ld	hl,emptyfcb	; use fcb for empty space
	ld	(params),hl
	ld	c,1
	call	findfst		; find first empty spot in directory
	call	ckfilpos	; check that one exists
	pop	hl
	ld	(params),hl	; restore original fcb
	ret	z		; done if no more space
	ex	de,hl
	ld	hl,15		; point to record count for this file
	add	hl,de
	ld	c,17		; and clear its space
	xor	a
getmt1:	ld	(hl),a
	inc	hl
	dec	c
	jp	nz,getmt1
	ld	hl,13
	add	hl,de
	ld	(hl),a		; also clear s1 byte
	call	chknmbr		; keep scratch1 within bounds
	call	fcbset		; write fcb entry to directory
	jp	sets2b7		; set bit 7 of s2 (unmodified)
;
; Close the current extent and open next one for reading.
;
getnext:
	xor	a
	ld	(closeflg),a	; clear close flag
	call	closeit		; close this extent
	call	ckfilpos
	ret	z		; done if not there
	ld	hl,(params)	; else get extent byte
	ld	bc,12
	add	hl,bc
	ld	a,(hl)
	inc	a		; bump extent byte
	and	1fh		; must be 0 - 31
	ld	(hl),a
	jp	z,gtnext1	; go if overflow
	ld	b,a
	ld	a,(extmask)	; else mask extent byte
	and	b
	ld	hl,closeflg	; check close flag
	and	(hl)
	jp	z,gtnext2	; if zero, read next extent
	jp	gtnext3		; else it's in memory
;
gtnext1:
	ld	bc,2
	add	hl,bc		; point to s2 byte
	inc	(hl)		; and increment it
	ld	a,(hl)
	and	0fh		; too many extents?
	jp	z,gtnext5	; if so, set error code
gtnext2:
	ld	c,15		; check first 15 bytes of fcb
	call	findfst		; find first match
	call	ckfilpos	; available?
	jp	nz,gtnext3	; go if so
	ld	a,(rdwrtflg)	; else check if reading (rdwrtflg = -1)
	inc	a
	jp	z,gtnext5	; go if reading (error)
	call	getempty	; else get an empty entry
	call	ckfilpos	; available?
	jp	z,gtnext5	; go if not (error)
	jp	gtnext4
;
gtnext3:
	call	openit1		; open this extent
gtnext4:
	call	strdata		; copy updated data (rec num, extent num, etc)
	xor	a
	jp	setstat		; clear status and return
;
gtnext5:
	call	ioerr1		; set error code
	jp	sets2b7		; clear bit 7 of s2 and return
;
; Read file sequentially
;
rdseq:	ld	a,1
	ld	(mode),a	; set sequential access mode
rdseq1:	ld	a,0ffh
	ld	(rdwrtflg),a	; don't read unwritten space
	call	strdata		; put rec num & ext num in fcb
	ld	a,(savnrec)	; get next record to read
	ld	hl,savnxt	; get number of records in extent
	cp	(hl)
	jp	c,rdseq2	; go if within this extent
	cp	128		; else is extent full?
	jp	nz,rdseq3	; go if not (end of file)
	call	getnext		; else open the next extent
	xor	a
	ld	(savnrec),a	; reset next record to read
	ld	a,(status)	; open successful?
	or	a
	jp	nz,rdseq3	; error if not
rdseq2:	call	comblk		; compute block number to read
	call	chkblk		; within bounds?
	jp	z,rdseq3	; go if not (error)
	call	logical		; convert blknum to logical sector
	call	trksec1		; set track and sector for this block
	call	doread		; and do the read
	jp	setnrec		; set next record to be accessed
;
rdseq3:	jp	ioerr1		; set read error and return
;
; Write file sequentially
;
wtseq:	ld	a,1
	ld	(mode),a	; set sequential access mode
wtseq1:	ld	a,0
	ld	(rdwrtflg),a	; allow opening empty extent
	call	chkwprt		; check write protect
	ld	hl,(params)	; get fcb pointer into HL
	call	ckrof1		; check file read only status
	call	strdata		; update fcb rec num & extent
	ld	a,(savnrec)	; get record number to write
	cp	128		; within range
	jp	nc,ioerr1	; error if not
	call	comblk		; compute block number
	call	chkblk		; check block number
	ld	c,0
	jp	nz,wtseq6	; go if block available
	call	getblock	; else get a new block number
	ld	(relblock),a	; save it
	ld	bc,0		; look for space from start
	or	a
	jp	z,wtseq2
	ld	c,a
	dec	bc		; get previous block number
	call	extblk		; so we can be close to it
	ld	b,h
	ld	c,l
wtseq2:	call	fndspace	; find nearest empty block
	ld	a,l		; is one available?
	or	h
	jp	nz,wtseq3	; go if so
	ld	a,2		; else out of space
	jp	setstat
;
wtseq3:	ld	(blknum),hl	; save block number
	ex	de,hl
	ld	hl,(params)	; update fcb
	ld	bc,16
	add	hl,bc
	ld	a,(bigdisk)	; block numbers 8 or 16 bit?
	or	a
	ld	a,(relblock)
	jp	z,wtseq4	; go if 16 bit
	call	adda2hl		; else point to proper entry
	ld	(hl),e		; and store new block number
	jp	wtseq5
;
wtseq4:	ld	c,a
	ld	b,0
	add	hl,bc		; point to 16 bit entry
	add	hl,bc
	ld	(hl),e		; and store new block number
	inc	hl
	ld	(hl),d
wtseq5:	ld	c,2		; flag writing to unused disk space
wtseq6:	ld	a,(status)	; ok so far?
	or	a
	ret	nz		; abort if not
	push	bc		; save write flag (C) for bios
	call	logical		; convert blknum to logical sector
	ld	a,(mode)	; get access mode
	dec	a		; 0 = random, 1 = sequential,
	dec	a		; 2 = directory write
	jp	nz,wtseq9	; go if not directory write
	pop	bc
	push	bc
	ld	a,c		; get write status
	dec	a
	dec	a
	jp	nz,wtseq9	; go if not directory write
	push	hl
	ld	hl,(dirbuf)	; else clear directory buffer
	ld	d,a
wtseq7:	ld	(hl),a
	inc	hl
	inc	d
	jp	p,wtseq7
	call	dirdma		; set DMA address for directory write
	ld	hl,(logsect)	; get start sector
	ld	c,2		; set flag to writing to unused space
wtseq8:	ld	(blknum),hl	; save sector to write
	push	bc
	call	trksec1		; get track and sector number
	pop	bc
	call	dowrite		; write out zeroed block
	ld	hl,(blknum)	; get sector number
	ld	c,0		; set write flag to normal
	ld	a,(blkmask)
	ld	b,a
	and	l
	cp	b		; entire block written?
	inc	hl
	jp	nz,wtseq8	; loop until block done
	pop	hl
	ld	(blknum),hl	; then set next sector number
	call	defdma		; reset DMA address
wtseq9:	call	trksec1		; get track and sector numbers
	pop	bc		; get write status flag
	push	bc
	call	dowrite		; write to disk
	pop	bc
	ld	a,(savnrec)	; get number of records in file
	ld	hl,savnxt	; and last record written
	cp	(hl)
	jp	c,wtseq10
	ld	(hl),a		; update record count
	inc	(hl)
	ld	c,2
wtseq10:
	nop			; patch to correct disk update when
	nop			; using blocking in the bios
	ld	hl,ccpbase
	push	af
	call	gets2		; set extent written flag
	and	7fh		; clear bit 7
	ld	(hl),a
	pop	af		; get record count for extent
	cp	127		; full?
	jp	nz,wtseq12	; go if not
	ld	a,(mode)	; in sequential mode?
	cp	1
	jp	nz,wtseq12	; go if not
	call	setnrec		; else set record number
	call	getnext		; and get next empty space in directory
	ld	hl,status
	ld	a,(hl)		; ok so far?
	or	a
	jp	nz,wtseq11	; go if not
	dec	a
	ld	(savnrec),a	; else set record count to -1
wtseq11:
	ld	(hl),0		; clear status
wtseq12:
	jp	setnrec		; set next record to access
;
; Set the fcb for the desired random record number based on the
; r0, r1, and r2 bytes. On entry C contains -1 if this is a read
; request, otherwise a new extent may be opened for writing.
;
position:
	xor	a
	ld	(mode),a	; set random I/O flag
positn1:
	push	bc
	ld	hl,(params)	; HL = fcb address
	ex	de,hl
	ld	hl,33
	add	hl,de
	ld	a,(hl)		; get r0 byte
	and	7fh		; bits 0-6 are record number
	push	af
	ld	a,(hl)		; get bit 7 of r0
	rla
	inc	hl
	ld	a,(hl)		; and bits 0-3 of r1 (extent byte)
	rla
	and	1fh
	ld	c,a		; and save in C
	ld	a,(hl)		; get extra extent byte
	rra
	rra
	rra
	rra
	and	0fh
	ld	b,a		; save extra extent in B
	pop	af		; recover record number
	inc	hl
	ld	l,(hl)		; check overflow byte r2
	inc	l		; check for disk space
	dec	l
	ld	l,6		; set error number
	jp	nz,positn5	; go if out of disk space
	ld	hl,32
	add	hl,de
	ld	(hl),a		; store record number in fcb
	ld	hl,12
	add	hl,de		; check extent byte
	ld	a,c
	sub	(hl)		; same extent as before?
	jp	nz,positn2	; go if not
	ld	hl,14
	add	hl,de		; else check extra extent s2
	ld	a,b
	sub	(hl)
	and	7fh
	jp	z,positn3	; go if also the same
positn2:
	push	bc		; another extent is required
	push	de
	call	closeit		; close current extent
	pop	de
	pop	bc
	ld	l,3		; set error number
	ld	a,(status)	; check status
	inc	a
	jp	z,positn4	; go if close error
	ld	hl,12
	add	hl,de
	ld	(hl),c		; else put extent into fcb
	ld	hl,14
	add	hl,de
	ld	(hl),b
	call	openit		; open the extent
	ld	a,(status)	; was it there?
	inc	a
	jp	nz,positn3	; go if so
	pop	bc		; else create a new one if writing
	push	bc
	ld	l,4		; set error number
	inc	c
	jp	z,positn4	; go if reading (unwritten space)
	call	getempty	; else try to find space
	ld	l,5		; set error number
	ld	a,(status)	; space available?
	inc	a
	jp	z,positn4	; go if not
positn3:
	pop	bc		; restore stack and clear error
	xor	a
	jp	setstat		; set status and return
;
positn4:
	push	hl		; set s2 byte to indicate error
	call	gets2		; HL points to s2
	ld	(hl),0c0h	; flag error
	pop	hl
positn5:
	pop	bc
	ld	a,l
	ld	(status),a	; set status saved in L
	jp	sets2b7
;
; Read a random record.
;
readran:
	ld	c,0ffh		; set read status
	call	position	; position to proper record
	call	z,rdseq1	; if no error, read it
	ret
;
; Write a random record
;
writeran:
	ld	c,0		; set write status
	call	position	; position to proper record
	call	z,wtseq1	; if no error, write it
	ret
;
; Compute random record number. On entry HL points to an fcb
; and DE contains relative location of a record. On exit C
; contains the r0 byte, B the r1 byte, and A the r2 byte. Z is
; set if the record is within bounds, otherwise an overflow
; occured.
;
comprand:
	ex	de,hl		; DE points to fcb
	add	hl,de		; HL points to relative position of record
	ld	c,(hl)		; get record number into BC
	ld	b,0
	ld	hl,12
	add	hl,de
	ld	a,(hl)		; get the extent
	rrca
	and	80h		; lower bit to bit 7 of A
	add	a,c		; add to record number
	ld	c,a
	ld	a,0
	adc	a,b
	ld	b,a
	ld	a,(hl)		; get upper bits of extent
	rrca			; into bits 0-3 of A
	and	0fh
	add	a,b		; add to r1 byte
	ld	b,a
	ld	hl,14
	add	hl,de
	ld	a,(hl)		; get the s2 (extra extent) byte
	add	a,a		; shift to bits 4-7
	add	a,a
	add	a,a
	add	a,a
	push	af		; save carry
	add	a,b		; add extra extent
	ld	b,a		; BC = record num + extent * 128
	push	af		; save carry
	pop	hl		; in bit 0 of L (overflow)
	ld	a,l
	pop	hl
	or	l
	and	1		; check carry flags
	ret
;
; Set up the fcb (r0, r1, r2) to the last record used for a file.
; Read the directory and compute the largest record number from
; each extent. The fcb r0, r1, r2 will be the maximum record number.
; This determines the space used by a random file.
;
ransize:
	ld	c,12
	call	findfst		; find first entry in directory
	ld	hl,(params)	; HL points to fcb
	ld	de,33		; point to r0 byte
	add	hl,de
	push	hl
	ld	(hl),d		; clear r0, r1, r2 (D = 0)
	inc	hl
	ld	(hl),d
	inc	hl
	ld	(hl),d
ransiz1:
	call	ckfilpos	; extent to process?
	jp	z,ransiz3	; done if not
	call	fcb2hl		; else set HL -> fcb in directory
	ld	de,15		; last record in extent
	call	comprand	; compute random record number
	pop	hl
	push	hl		; check against fcb
	ld	e,a
	ld	a,c
	sub	(hl)		; Carry set if record in fcb is
	inc	hl		; larger than this extant
	ld	a,b
	sbc	a,(hl)
	inc	hl
	ld	a,e
	sbc	a,(hl)
	jp	c,ransiz2	; go if fcb record is larger
	ld	(hl),e		; else put these values into fcb
	dec	hl
	ld	(hl),b
	dec	hl
	ld	(hl),c
ransiz2:
	call	findnxt		; get next extent
	jp	ransiz1		; loop til done
;
ransiz3:
	pop	hl
	ret
;
; BDOS function 36 - set random record
; Set fcb random record position to last record used
;
settran:
	ld	hl,(params)	; HL -> fcb
	ld	de,32		; last record used
	call	comprand	; compute random position
	ld	hl,33
	add	hl,de		; move rand pos vals into fcb
	ld	(hl),c		; C -> r0
	inc	hl
	ld	(hl),b		; B -> r1
	inc	hl
	ld	(hl),a		; A -> r2
	ret
;
; Select active drive and update login vector and bitmap table.
;
logindrv:
	ld	hl,(login)	; get login vector
	ld	a,(active)	; get default drive
	ld	c,a
	call	shiftr		; position bit for this drive
	push	hl
	ex	de,hl
	call	select		; select this drive
	pop	hl
	call	z,slcterr	; error if not valid drive
	ld	a,l
	rra
	ret	c		; done if already active
	ld	hl,(login)	; else update login vector
	ld	c,l
	ld	b,h
	call	setbit
	ld	(login),hl
	jp	bitmap		; update bit map
;
; BDOS function 14 - select disk
; Set active disk number
;
setdsk:	ld	a,(eparam)	; get bdos function parameter
	ld	hl,active
	cp	(hl)		; same drive?
	ret	z		; done if so
	ld	(hl),a		; else log it in
	jp	logindrv
;
; Select drive based on drive spec in fcb.
;
autosel:
	ld	a,0ffh
	ld	(auto),a	; set auto select flag
	ld	hl,(params)	; HL -> fcb
	ld	a,(hl)		; get drive spec
	and	1fh
	dec	a		; drive number
	ld	(eparam),a	; save for select routine
	cp	1eh
	jp	nc,autosl1	; don't change if drive was 31 (?)
	ld	a,(active)
	ld	(olddrv),a	; save currently active drive
	ld	a,(hl)
	ld	(autoflag),a	; save drive spec from fcb
	and	0e0h
	ld	(hl),a		; remove bits from spec (?)
	call	setdsk		; select and log in drive
autosl1:
	ld	a,(userno)	; get user number
	ld	hl,(params)	; add to drive spec
	or	(hl)
	ld	(hl),a
	ret
;
; BDOS function 12 - return version number
;
getver:	ld	a,22h		; version 2.2
	jp	setstat
;
; BDOS function 13 - reset disk system
;
rstdsk:	ld	hl,0
	ld	(wrtprt),hl	; clear write protect status
	ld	(login),hl	; and login vector
	xor	a
	ld	(active),a	; select drive A
	ld	hl,defaultdma
	ld	(userdma),hl	; set default dma address
	call	defdma
	jp	logindrv	; log in drive A
;
; BDOS function 15 - open file
;
openfile:
	call	clears2		; clear s2 byte
	call	autosel		; select disk
	jp	openit		; go open file
;
; BDOS function 16 - close file
;
closefile:
	call	autosel		; select disk
	jp	closeit		; go close file
;
; BDOS function 17 - search for first
; Get first entry for file in fcb
;
getfirst:
	ld	c,0
	ex	de,hl
	ld	a,(hl)		; if first byte is a wildcard,
	cp	'?'		; get the first entry
	jp	z,getfst1	; go if first byte is '?'
	call	setext
	ld	a,(hl)		; get extension byte from fcb
	cp	'?'		; if it's a wildcard,
	call	nz,clears2	; clear s2 byte
	call	autosel		; select drive
	ld	c,15
getfst1:
	call	findfst		; find first entry
	jp	movedir		; and put in user's dma space
;
; BDOS function 18 - search for next
;
getnxt:	ld	hl,(savefcb)	; restore pointer to fcb
	ld	(params),hl
	call	autosel		; select drive
	call	findnxt		; find next entry
	jp	movedir		; and put in user's dma space
;
; BDOS function 19 - delete file
;
delfile:
	call	autosel		; select drive
	call	erafile		; remove the file
	jp	ststatus	; set status and return
;
; BDOS function 20 - read sequential
;
readseq:
	call	autosel		; select drive
	jp	rdseq		; go do read
;
; BDOS function 21 - write sequential
;
writeseq:
	call	autosel		; select drive
	jp	wtseq		; go do write
;
; BDOS function 22 - make file
;
fcreate:
	call	clears2		; clear s2 byte
	call	autosel		; select drive
	jp	getempty	; find empty directory space
;
; BDOS function 23 - rename file
;
renfile:
	call	autosel		; select drive
	call	chgnames	; change the file name
	jp	ststatus	; set status and return
;
; BDOS function 24 - return login vector
;
getlog:	ld	hl,(login)
	jp	getprm1
;
; BDOS function 25 - return current disk
;
getcrnt:
	ld	a,(active)
	jp	setstat
;
; BDOS function 26 - set DMA address
;
putdma:	ex	de,hl
	ld	(userdma),hl	; save address
	jp	defdma		; set in BIOS also
;
; BDOS function 27 - get addr (alloc)
;
getaloc:
	ld	hl,(allocvect)
	jp	getprm1
;
; BDOS function 29 - get R/O vector
;
getrov:	ld	hl,(wrtprt)
	jp	getprm1
;
; BDOS function 30 - set file attributes
;
setattr:
	call	autosel		; select drive
	call	saveattr	; set attributes
	jp	ststatus	; set status and return
;
; BDOS function 31 - get addr (disk parms)
;
getparm:
	ld	hl,(diskpb)
getprm1:
	ld	(status),hl
	ret
;
; BDOS function 32 - set/get user code
;
getuser:
	ld	a,(eparam)
	cp	0ffh		; if E was -1, return user number
	jp	nz,setuser	; else set it to value in E
	ld	a,(userno)
	jp	setstat
;
setuser:
	and	1fh
	ld	(userno),a
	ret
;
; BDOS function 33 - read random
;
rdrandom:
	call	autosel		; select drive
	jp	readran		; go do read
;
; BDOS function 34 - write random
;
wrrandom:
	call	autosel		; select drive
	jp	writeran	; go do write
;
; BDOS function 35 - compute file size
;
filesize:
	call	autosel		; select drive
	jp	ransize		; go compute size
;
; BDOS function 37 - reset drive
; On entry DE bits are set for the drives to be logged off.
;
logoff:	ld	hl,(params)	; get drives to log off
	ld	a,l
	cpl			; invert bit patter
	ld	e,a
	ld	a,h
	cpl
	ld	hl,(login)	; get login vector
	and	h
	ld	d,a		; clear login vector bits
	ld	a,l
	and	e
	ld	e,a
	ld	hl,(wrtprt)	; get write protect vector
	ex	de,hl
	ld	(login),hl	; update login vector
	ld	a,l
	and	e		; clear write protect vector bits
	ld	l,a
	ld	a,h
	and	d
	ld	h,a
	ld	(wrtprt),hl	; update write protect vector
	ret
;
; This is called to return to the user's code.
;
goback:	ld	a,(auto)	; test if drive auto selected
	or	a
	jp	z,goback1	; go if not
	ld	hl,(params)	; else reset drive spec in fcb
	ld	(hl),0
	ld	a,(autoflag)	; was a change made?
	or	a
	jp	z,goback1	; go if not
	ld	(hl),a
	ld	a,(olddrv)	; else select the old drive
	ld	(eparam),a
	call	setdsk
goback1:
	ld	hl,(usrstack)	; set stack pointer to user's stack
	ld	sp,hl
	ld	hl,(status)	; set return status
	ld	a,l
	ld	b,h
	ret
;
; BDOS function 40 - write random with zero fill
;
wtspecl:
	call	autosel		; select drive
	ld	a,2
	ld	(mode),a	; set write mode
	ld	c,0
	call	positn1		; position the file
	call	z,wtseq1	; if no error, then do write
	ret
;
; BDOS data
;
emptyfcb:	db	0e5h	; empty directory fcb
;
wrtprt:		dw	0	; write protect status for 16 drives
login:		dw	0	; login bits for 16 drives
userdma:	dw	defaultdma	; user's dma address
;
; Scratch area for parameters
;
scratch1:	dw	0	; position within directory
scratch2:	dw	0	; last selected track
scratch3:	dw	0	; last selected sector
;
dirbuf:		dw	0	; address of directory buffer to use
diskpb:		dw	0	; address of dpb to use
chkvect:	dw	0	; address of check vector
allocvect:	dw	0	; address of bit map
;
; DPB from the BIOS
;
sectors:	dw	0	; sectors per track
blkshft:	db	0	; block shift
blkmask:	db	0	; block mask
extmask:	db	0	; extent mask
dsksize:	dw	0	; disk size (number of blocks - 1)
dirsize:	dw	0	; directory size
alloc0:		dw	0	; bit map (used space)
alloc1:		dw	0
offset:		dw	0	; first track to use
xlate:		dw	0	; address of sector translate table
;
closeflg:	db	0	; close flag (-1 if extent written)
rdwrtflg:	db	0	; read/write flag (-1 = read, 0 = write)
fndstat:	db	0	; file found status
mode:		db	0	; I/O mode (0 = random, 1 = sequential, 2 = write directory)
eparam:		db	0	; value of E on BDOS entry
relblock:	db	0	; position in fcb of block number
counter:	db	0	; byte counter for directory name search
savefcb:	dw	0	; save fcb address for directory search
		dw	0
;
bigdisk:	db	0	; 0 = disk > 256 blocks
auto:		db	0	; if non-zero, auto select drive
olddrv:		db	0	; save current drive when auto select
autoflag:	db	0	; non-zero if auto select changed drives
savnxt:		db	0	; next record number to access
savext:		db	0	; extent number of file
savnrec:	dw	0	; number of records in file
blknum:		dw	0	; block number (physical sector) in file
logsect:	dw	0	; starting sector in a block
fcbpos:		db	0	; position in directory of fcb
filepos:	dw	0	; file position within directory
;
cksumtbl:	db	0,0,0,0,0,0,0,0	; directory check sum
		db	0,0,0,0,0,0,0,0	; bytes for 16 drives
;
		db	0,0,0,0	; filler to end of record
;
	end
;

