;
; Montezuma Micro CP/M 2.32 BIOS
; BIOS - Basic Input/Output System
; for Radio Shack Model 4 computer.
;
; This source code is for the Z80ASM assembler
; from SLR systems.
;
	include	system.inc	; get system definitons
;
biosvec		equ	0
bdosentry	equ	bdosbase+06h
bdosvec		equ	5
ccp		equ	ccpbase
ccpentry	equ	ccpbase+03h
defaultdrive	equ	4
;
dirbuf		equ	biosbase+0f00h		; 0f900h
dskiobfr	equ	biosbase+0f80h		; 0f980h
;
iobyte		equ	3
lastmemloc	equ	0ffffh
lastvidline	equ	0ff30h
videobase	equ	0f800h
nmiptr		equ	67h
nmivec		equ	66h
tpa		equ	100h
;
; hardware ports
;
optreg		equ	84h	; option register
soundport	equ	90h	; sound option
nmireg		equ	0e4h
modemstatus	equ	0e8h
rstenable	equ	0e8h
brgen		equ	0e9h	; baud rate generator
sioc		equ	0eah	; uart control & status
siod		equ	0ebh	; uart data register
fcsr		equ	0f0h	; fdc control & status
ftrk		equ	0f1h	; fdc track register
fsec		equ	0f2h	; fdc sector register
fdat		equ	0f3h	; fdc data register
fsel		equ	0f4h	; floppy select port
prtport		equ	0f8h	; printer port
;
prtstatmask	equ	0f0h	; printer status mask
;
; hardware addresses
;
keyboardcol0	equ	0f401h
kybdsplmask	equ	0f47fh
;
; miscellaneous definitions
;
lf		equ	0ah
formfeed	equ	0ch
cr		equ	0dh
jumpcode	equ	0c3h
retcode		equ	0c9h
;
; The BIOS
;
	org	biosbase
;
bios:	jp	boot
wbootvec:
	jp	wboot
	jp	const
	jp	conin
	jp	conout
	jp	list
	jp	punch
	jp	reader
	jp	home
	jp	seldsk
	jp	settrk
	jp	setsec
	jp	setdma
	jp	read
	jp	write
	jp	listst
	jp	sectran
;
defaultiobyte:	db	81h
signonflag:	db	0bh	; show banner on boot if not zero
;
; accessed by the config utility
;
numdrives:	db	4
biosversion:	db	23h
;
; These pointers are here for the benefit of the config utility
;
	dw	dph0ptr
	dw	dpb0vec
	dw	dpb1vec
	dw	dpb2vec
	dw	dpb3vec
	dw	constsiovec
	dw	kybdchar
	dw	videochar
	dw	lptchar
	dw	setbaudvec
;
boot:	ld	sp,0
	call	initram
	xor	a
	ld	(defaultdrive),a	; set boot disk to A
	ld	(ramdrvflag),a	; disable ram drive
	dec	a
	ld	(dsk0track),a	; set all drives to track 0
	ld	(dsk1track),a
	ld	(dsk2track),a
	ld	(dsk3track),a
	ld	hl,dph0
	ld	bc,16		; each dph is 16 bytes
	ld	(dph0ptr),hl	; init pointers to dph's
	add	hl,bc
	ld	(dph1ptr),hl
	add	hl,bc
	ld	(dph2ptr),hl
	add	hl,bc
	ld	(dph3ptr),hl
	ld	hl,0
	ld	a,0efh
	out	(optreg),a
	ld	(hl),3ch	; test for 128K ram
	ld	a,8fh
	out	(optreg),a
	ld	a,(hl)
	ld	(hl),0c3h
	cp	(hl)
	jr	nz,boot1	; go if no extended ram
	ld	a,0efh
	out	(optreg),a
	call	zeroramdisk
	add	hl,hl
	ld	a,0ffh
	out	(optreg),a
	call	zeroramdisk
	ld	a,cr
	ld	(ramdrvflag),a	; ram disk enabled
	ld	hl,dph4
	ld	(dph4ptr),hl	; set up dph for ram disk
	ld	a,8fh
	out	(optreg),a
boot1:	ld	hl,signon
	ld	a,(signonflag)	; signonflag can only be cleared by the config utility
	or	a
	call	nz,printstr
	ld	c,0
	jp	ccp
;
; This routine "formats" the ram drive
;
zeroramdisk:
	ld	(hl),0e5h	; fill ram drive with erased character
	inc	hl
	bit	7,h		; until we hit the end of 32K bank
	ret	nz
	jr	zeroramdisk
;
signon:	db	1ah,7,0fh
	db	'TRS-80 Model 4 '
;	db	'64'
	db	(MEMORY/10)+30h
	db	(MEMORY MOD 10)+30h
	db	'k CP/M vers 2.2 (c) (p) 1982 Digital Research Inc.'
	db	15h,cr,lf
	db	'BIOS rel. 2.32 (c) (p) 1984 Montezuma Micro/JBO'
	db	15h,0eh,cr,lf,lf
ramdrvflag:	db	cr	; will be set to null if no ramdrive present
	db	'>>> Memory Drive M: ',16h,'ENABLED'
	db	16h,cr,lf,lf,0
;
wboot:	ld	sp,tpa
	call	flush
	call	initsys
	ld	c,0		; select drive A
	call	seldsk
	ld	bc,10
	add	hl,bc		; point to dpb entry in dph
	ld	a,(hl)
	inc	hl
	ld	h,(hl)
	ld	l,a		; hl -> dpb
	ld	a,(hl)
	inc	hl
	ld	h,(hl)
	ld	l,a		; hl = spt
	ld	(currspt),hl
	ld	hl,0
	ld	(currtrk),hl
	ld	l,2
	ld	(currsec),hl
	ld	hl,ccp
	ld	(dmaptr),hl
;
; Read in CCP and BDOS
;
	ld	b,44		; number of sectors for CCP & BDOS
wboot1:	push	bc
	ld	bc,(currtrk)
	call	settrk
	ld	bc,(currsec)
	call	setsec
	call	read
	or	a
	jr	nz,wboot
	ld	hl,currtrk
	call	incsector
	ld	hl,(dmaptr)
	ld	bc,128
	add	hl,bc
	ld	(dmaptr),hl
	pop	bc
	djnz	wboot1
	ld	a,(defaultdrive)
	and	0fh
	ld	c,a
	call	seldsk
	ld	a,h
	or	l
	jr	nz,wboot2
	ld	(defaultdrive),a
wboot2:	ld	a,(defaultdrive)
	ld	c,a
	jp	ccpentry
;
; Access to I/O devices
;
; Note that each routine calls tablejump, which accesses
; a table of four pointers to pointers following the call
; to tablejump. In three cases (const, conin, and conout)
; the third entry is not a pointer to a pointer. The third
; entry corresponds to the CON device being set to BAT:
; (in = RDR, out = LST). In all three cases, this is
; detected by the routine before calling tablejump, so
; this error is not a problem.
;
; It appears to be done this way so that if CON is set to
; BAT, and RDR & LST are configured to be other devices,
; CON requests will go to those other devices.
;
const:	call	getconiob
	jr	z,constrdr	; go if console = lpt
	call	tablejump
;
	dw	constsiovec
	dw	constreadervec
	dw	constbat	; this should be a pointer to a pointer!
	dw	constuc1vec
;
; Get reader field of iobyte
;
constrdr:
	ld	a,(iobyte)
	rrca
	rrca
	and	3
	call	tablejump
;
	dw	constsiovec
	dw	constrdrkbvec
	dw	constrdrbatvec
	dw	constrdruc1vec
;
conin:	call	getconiob
	jr	z,reader	; go if console = lpt
	call	tablejump
;
	dw	coninsiovec
	dw	coninkbdvec
	dw	coninrdr	; this should be a pointer to a pointer!
	dw	coninuc1vec
;
conout:	call	getconiob
	jr	z,list		; go if console = lpt
	call	tablejump
;
	dw	conoutsiovec
	dw	conoutcrtvec
	dw	conoutlst	; this should be a pointer to a pointer!
	dw	conoutuc1vec
;
; Get iobyte console field
; return with Z set if con = lpt
;
getconiob:
	ld	a,(iobyte)
	and	3
	cp	2
	ret
;
list:	ld	a,(iobyte)
	rlca
	rlca
	and	3
	call	tablejump
;
	dw	conoutsiovec
	dw	conoutcrtvec
	dw	conoutlptvec
	dw	conoutul1vec
;
listst:	ld	a,(iobyte)
	rlca
	rlca
	and	3
	call	tablejump
;
	dw	liststsiovec
	dw	liststcrtvec
	dw	liststlptvec
	dw	liststul1vec
;
punch:	ld	a,(iobyte)
	rlca
	rlca
	rlca
	rlca
	and	3
	call	tablejump
;
	dw	conoutsiovec
	dw	punchvidvec
	dw	punchup1vec
	dw	punchup2vec
;
reader:	ld	a,(iobyte)
	rrca
	rrca
	and	3
	call	tablejump
;
	dw	coninsiovec
	dw	readerkbvec
	dw	readerur1vec
	dw	readerur2vec
;
; Indirect jump through vector table
; On entry the table address is on the stack
; and A contains the entry number (iobyte bits).
;
tablejump:
	pop	hl	; get table address off stack
	add	a,a	; compute offset into table
	ld	e,a
	ld	d,0
	add	hl,de	; get address of entry vector
	ld	e,(hl)
	inc	hl
	ld	d,(hl)	; DE = entry vector
	ex	de,hl	; HL = entry vector
	ld	e,(hl)
	inc	hl
	ld	d,(hl)	; DE = function address
	ex	de,hl	; put in HL for jump
	jp	(hl)	; jump to function from table
;
; Vector Table for I/O
; Every entry in this table should be referenced.
; It is unknown why there are unreferenced entries.
;
constsiovec:	dw	constsio
coninsiovec:	dw	sioinput
conoutsiovec:	dw	siooutput
liststsiovec:	dw	sioxtst
constreadervec:	dw	constkbd
coninkbdvec:	dw	coninkbd
conoutcrtvec:	dw	conoutcrt
liststcrtvec:	dw	coninuc1
constuc1vec:	dw	constkbd
coninuc1vec:	dw	coninkbd
conoutuc1vec:	dw	conoutcrt
		dw	coninuc1
		dw	constbat
		dw	coninrdr
conoutlptvec:	dw	prtout
liststlptvec:	dw	prtstat
		dw	constbat
		dw	coninrdr
conoutul1vec:	dw	prtout
liststul1vec:	dw	prtstat
constrdrkbvec:	dw	constkbd
readerkbvec:	dw	coninkbd
		dw	conoutlst
		dw	coninuc1
constrdrbatvec:	dw	constsio
readerur1vec:	dw	sioinput
		dw	siooutput
		dw	sioxtst
constrdruc1vec:	dw	constsio
readerur2vec:	dw	sioinput
		dw	siooutput
		dw	sioxtst
		dw	constbat
		dw	coninrdr
punchvidvec:	dw	conoutcrt
		dw	coninuc1
		dw	constsio
		dw	sioinput
punchup1vec:	dw	siooutput
		dw	sioxtst
		dw	constsio
		dw	sioinput
punchup2vec:	dw	siooutput
		dw	sioxtst
;
; End of I/O vector table
;
constbat:	xor	a
conoutlst:	ret
;
coninrdr:	ld	a,1ah
		ret
;
coninuc1:	ld	a,0ffh
		ret
;
; Initialize ram. Called by boot.
;
initram:
	call	clearmem
	dw	exbios
	dw	23h	; clear from exbios to dph pointers
;
	call	initvideodata
	call	setbaudrate
	ld	a,(defaultiobyte)
	ld	(iobyte),a
	ld	a,retcode
	ld	(exbios),a
;
; Initialize system. Called by wboot.
;
initsys:
	call	clearmem
	dw	seekdisk
	dw	23h	; clear from currentdisk to alv0
;
	call	initkybdflags
	call	initprtflags
	ld	a,jumpcode
	ld	(biosvec),a
	ld	(bdosvec),a
	ld	hl,wbootvec
	ld	(biosvec+1),hl
	ld	hl,bdosentry
	ld	(bdosvec+1),hl
	ret
;
; Output string pointed to by HL to console
;
printstr:
	ld	a,(hl)
	inc	hl
	or	a
	ret	z
	push	hl
	ld	c,a
	call	conout
	pop	hl
	jr	printstr
;
; Delay number of msec in A
;
delayms:
	push	bc
dlyms1:	ld	b,99h
dlyms2:	djnz	dlyms2
	ld	b,9ah
dlyms3:	djnz	dlyms3
	nop
	dec	a
	jr	nz,dlyms1
	pop	bc
	ret
;
; Clear memory. On entry (sp) = address (sp+2) = byte count.
;
clearmem:
	xor	a
	ex	(sp),hl
	ld	e,(hl)	; fetch destination address
	inc	hl
	ld	d,(hl)
	inc	hl
	ld	c,(hl)	; fetch byte count
	inc	hl
	ld	b,(hl)
	inc	hl
	ex	(sp),hl	; put return adrs to caller on stack
;
; Fill memory at HL with byte in A for BC bytes
;
fillmem:
	ld	h,d
	ld	l,e
	ld	(hl),a	; clear first byte of destination
	dec	bc
	ld	a,b
	or	c
	ret	z
	inc	de	; if byte count not zero,
	ldir		; clear remaining block
	ret
;
; Increment 16 bits at (hl)
;
inc16hl:
	inc	(hl)
	ret	nz
	inc	hl
	inc	(hl)
	dec	hl
	ret
;
; Compare 16 bits at (hl) & (de)
;  Return Z if equal
;
cmphlde:
	inc	de
	inc	hl
	ld	a,(de)
	cp	(hl)
	dec	de
	dec	hl
	ret	nz
	ld	a,(de)
	cp	(hl)
	ret
;
; Compare B bytes at (hl) & (de)
; Return Z if all are equal
;
cmpbytes:
	ld	a,(de)
	inc	de
	cp	(hl)
	inc	hl
	ret	nz
	djnz	cmpbytes
	ret
;
constkbd:
	ld	a,(kybdchar)
	or	a
	jr	nz,constkbd1
	call	kbscan
	ret	z
	ld	(kybdchar),a
constkbd1:
	or	0ffh
	ret
;
coninkbd:
	ld	hl,kybdchar
	ld	a,(hl)
	ld	(hl),0
	or	a
	ret	nz
coninkbd1:
	call	kbscan
	jr	z,coninkbd1
	ret
;
; Scan keyboard and return ascii decoded key pressed (if any)
; Keyboard mapped from 0f400h-0f7ffh
;
; Bit                      Address
;      f401  f402  f404  f408  f410  f420  f440  f480
; 0     @     H     P     X     0     8    enter lshift
; 1     A     I     Q     Y     1     9    clear rshift
; 2     B     J     R     Z     2     :    break  ctl
; 3     C     K     S           3     ;     up    cap
; 4     D     L     T           4     ,    down   F1
; 5     E     M     U           5     -    left   F2
; 6     F     N     V           6     .    right  F3
; 7     G     O     W           7     /    space
;
; The lower 8 bits of the address select one or more columns,
; and the data read from that address will have a bit set if any
; key in that row is pressed. For example:
;
; If address 0f47fh is read and the data is 10h, then one of the
; keys in row 4 is pressed (D, L, T, 4, ',', or down arrow),
; excluding F1. Then read the addresses 0f401h, 0f402h, 0f404h,
; 0f410h, 0f420h, and 0f440h to determine which column the key is in.
;
; Reading 0f4ffh will return non-zero if any key is pressed.
;
kbscan:	ld	a,8eh		; map keyboard to 0f400h-0f7ffh
	out	(optreg),a
	call	scanfk		; check for function keys
	jp	nz,kbscandone	; done if SHIFT|CTL|CAP|F1|F2|F3 key
kbscan1:
	ld	de,keyboardcol0	; else check for regular keys
	ld	b,0
	ld	hl,keycoldata
;
; this will loop through all columns
; except SHIFT, CTL, CAP, F1-F3
;
kbscan2:
	ld	a,(de)		; get raw key data
	ld	c,a
	xor	(hl)		; check for change
	ld	(hl),c		; save new raw key data
	and	c
	jr	nz,kbscan5	; go if change
	inc	b		; else incr col number
	inc	hl		; incr col data pointer
	rlc	e		; incr column
	jp	p,kbscan2	; loop if more columns to do
	ld	a,(lastkeycoladrs)
	ld	e,a
	ld	a,(de)
	ld	c,a
kbscan3:
	ld	hl,(lastcoldataptr)
	ld	a,(hl)
	and	c
	jr	nz,kbscan4
	ld	hl,0800h	; set initial delay before repeat
	ld	(kbrptcnt),hl
	jr	kbscandone
;
kbscan4:
	ex	de,hl
	ld	hl,(kbrptcnt)
	dec	hl
	ld	(kbrptcnt),hl
	ld	a,h
	or	l
	ld	a,0
	jr	nz,kbscandone	; done if repeat not timed out
	ld	(de),a
	ld	hl,(kbrptreload)
	ld	(kbrptcnt),hl
	jr	kbscan1		; rescan keys
;
; change has occurred in scanned column
;
kbscan5:
	ld	c,a
	call	debounce
	jr	z,kbscandone	; go if no key press after debounce
	ld	a,e		; else save col & data pointers
	ld	(lastkeycoladrs),a
	ld	(lastcoldataptr),hl
	sla	b		; mult col number by 8 (8 char/col)
	sla	b
	sla	b
	dec	b
kbscan6:
	inc	b
	srl	c		; shift row number to get offset
	jr	nc,kbscan6	; shift for each row
	ld	hl,kybdflags
	ld	a,b		; get col num * 8
	cp	20h
	jr	nc,kbscan7	; go if not alpha key or '@'
	bit	2,(hl)		; test control bit in flags
	jr	nz,kbscandone	; go if control character
	set	6,b		; else make base ascii char
	or	a		; is column 0?
	jr	z,kbscan9	; if so, code in B is ascii char
	ld	a,(capslock)	; test caps lock flag
	or	a
	jr	nz,kbscan9	; go if caps lock on
	set	5,b		; else make lower case
	ld	a,3
	and	(hl)		; test shift key flags
	jr	z,kbscan9	; go if neither shift flag set
	ld	a,20h		; change character case
	xor	b
	jr	kbscandone
;
; Key is not alpha or '@'
;
kbscan7:
	sub	20h		; adjust offset to table offset
	ld	c,a
	ld	b,0
	ld	de,keytable
	ex	de,hl
	add	hl,bc		; add offset to table address
	ld	bc,24
	ld	a,(de)		; get kybdflags
	and	7		; mask LSHIFT, RSHIFT, & CTL bits
	jr	z,kbscan8	; go if none of those bits set
	add	hl,bc		; else adjust to shift key table
	and	4
	jr	z,kbscan8	; go if control flag not set
	add	hl,bc		; else adjust to control key table
kbscan8:
	ld	b,(hl)		; get ascii code for character
kbscan9:
	ld	a,b
kbscandone:
	ld	c,a		; save decoded key
	ld	a,8fh		; unmap keyboard
	out	(optreg),a
	ld	a,c		; restore decoded key
	or	a
	ret
;
; Check for LSHIFT, RSHIFT, CTL, CAP, F1-F3 keys
;
scanfk:	ld	hl,(Fkeypointer)
	ld	a,h
	or	l
	jr	nz,scanfk5	; go if not first scan
	ld	de,kybdsplmask	; else check for any key except SHIFT, CTL, CAP, F1, F2, F3
	ld	c,e		; mask for SHIFT, CTL, CAP, F1-F3
	ld	a,(de)
	or	a
	jr	z,scanfk1	; go if no non-special key pressed
	ld	c,7		; else set mask for LSHIFT, RSHIFT, CTL only
scanfk1:
	ld	hl,kybdflags
	inc	e		; check for SHIFT, CTL, CAP, F1, F2, F3
	ld	a,(de)		; get raw key data from column 7
	and	c		; mask bits of interest
	ld	c,a
	xor	(hl)		; changes from last time
	ld	(hl),c		; set new flag bits
	and	c		; any special key just pressed?
	ret	z		; done if not
	ld	c,a
	call	debounce	; delay 15ms and refetch raw data
	ret	z		; done if no key press
	bit	3,c		; is it CAPS key?
	jr	z,scanfk2	; go if not
	ld	a,(capslock)	; else toggle caps lock
	xor	1
	ld	(capslock),a
	push	bc		; beep if caps lock went on
	ld	c,28h
	call	nz,bell1
	pop	bc
scanfk2:
	ld	a,70h		; test for function keys
	and	c
	ret	z		; done if not a function key
	rlca			; else mult by 2
	ex	de,hl		; DE points to kybdflags
	ld	hl,F1shifttbl	; HL points to function key translation tables
	ld	bc,10
scanfk3:
	sbc	hl,bc		; adjust pointer for correct F key
	rlca
	jr	nc,scanfk3
	ld	c,30
	ld	a,(de)		; get kybdflags
	and	3
	jr	nz,scanfk4	; go if shift key set
	ld	a,(de)
	and	4
	jr	z,scanfk5	; go if ctrl key not set
	add	hl,bc		; adjust for ctrl F key
scanfk4:
	add	hl,bc		; adjust for shift F key
scanfk5:
	ld	a,(hl)		; get translated code for function key
	inc	hl
	or	a
	jr	z,scanfk6
	ld	(Fkeypointer),hl
	ret	p
scanfk6:
	ld	hl,0
	ld	(Fkeypointer),hl
	and	7fh
	ret
;
; Delay, then re-read column data
;
debounce:
	ld	a,0fh
	call	delayms		; delay 15 msec
	ld	a,(de)
	and	c
	ret
;
initkybdflags:
	call	clearmem
	dw	kybdchar
	dw	0eh		; clear from kybdchar to kbrptcnt
;
	ret
;
kybdchar:	db	0	; last character from keyboard scan
Fkeypointer:	dw	0
;
; Raw data for key columns
;
keycoldata:	db	0,0,0
lastcoldata:	db	2,0,0,0
;
; Keyboard flags. Bits are set when corresponding keys are active:
; bit 0 = left shift
; bit 1 = right shift
; bit 2 = control bit
; bit 3 = caps
; bit 4 = F1
; bit 5 = F2
; bit 6 = F3
;
kybdflags:	db	0
lastkeycoladrs:	db	8
lastcoldataptr:	dw	lastcoldata
kbrptcnt:	dw	7d8h	; keyboard repeat counter
kbrptreload:	dw	80h	; keyboard repeat factor (default = 128)
capslock:	db	0	; will be non-zero when caps lock is on
;
; Key table entries for numeric and special keys
;
keytable:	db	'0123456789:;,-./'
		db	cr,1bh,3,0bh,lf,8,9,' '
;
; Shifted key entries
;
keytables:	db	'0!"#$%&',27h,'()*+<=>?'
		db	cr,1bh,3,0bh,lf,8,9,' '
;
; Control key entries
;
keytablec:	db	'0|2~4^6`[]:;{_}\'
		db	cr,7fh,3,0bh,lf,8,9,' '
;
; Key table entries for function keys.
; Up to 10 bytes for each key translation.
; Each sequence is null terminated.
;
F1transtbl:	db	'DIR A:',cr,0,0,0
F2transtbl:	db	'DIR B:',cr,0,0,0
F3transtbl:	db	'DIR M:',cr,0,0,0
F1shifttbl:	db	'PIP A:=',0,0,0
F2shifttbl:	db	'PIP B:=',0,0,0
F3shifttbl:	db	'PIP M:=',0,0,0
F1ctrltbl:	db	'DUP',cr,0,0,0,0,0,0
F2ctrltbl:	db	'CONFIG',cr,0,0,0
F3ctrltbl:	db	'STAT ',0,0,0,0,0
;
; Video output routine
; On entry C = character to output
;
conoutcrt:
	ld	a,8eh
	out	(optreg),a
	ld	a,(videochar)
	call	putvchar
	call	outputchar
	call	getvchar
	ld	(videochar),a
	ld	hl,cursorflag
	xor	(hl)
	call	putvchar
	ld	a,8fh
	out	(optreg),a
	ret
;
getvchar:
	call	getcurslocation
	ld	a,(hl)
	ret
;
putvchar:
	call	getcurslocation
	ld	(hl),a
	ret
;
; On exit HL = videobase + videopos * 80
; HL -> character position on screen
;
getcurslocation:
	ld	hl,(videopos)
getcursadrs:
	push	bc
	push	de
	ld	bc,videobase
	ld	d,c
	ld	e,l
	ld	c,h
	ld	h,d
	add	hl,hl
	add	hl,hl
	add	hl,de
	add	hl,hl
	add	hl,hl
	add	hl,hl
	add	hl,hl
	add	hl,bc
	pop	de
	pop	bc
	ret
;
outputchar:
	ld	a,(escflag)	; in escape sequence?
	or	a
	jr	nz,escseq	; go if esc sequence
	ld	a,c		; get char to put on screen
	cp	20h
	jr	c,outctrl	; go if control code
	ld	a,(videorevflag)
	or	c
	call	putvchar
	ld	hl,(videopos)
	jp	inccursor
;
; Process escape sequence
; On entry A = sequence state
;
escseq:	cp	2
	jr	c,escseq2	; go if last character of sequence
	jr	z,escseq4	; go if second char of seq
	ld	a,c		; else is first char after esc
	cp	3dh		; is it cursor position code?
	ld	a,2
	jr	z,escseq6	; go if cursor position
	call	cursoronoff	; else is cursor on or off code
	xor	a		; clear esc sequence
	jr	escseq6
;
; last character of esc sequence = col
;
escseq2:
	ld	a,(cursposrow)
	ld	l,a
	ld	a,c
	sub	20h
	cp	80
	jr	c,escseq3
	ld	a,4fh
escseq3:
	ld	h,a
	ld	(videopos),hl
	xor	a
	jr	escseq6
;
; second character of esc sequence = row
;
escseq4:
	ld	a,c
	sub	20h
	cp	24
	jr	c,escseq5
	ld	a,23
escseq5:
	ld	(cursposrow),a
	ld	a,1
escseq6:
	ld	(escflag),a
	ret
;
outctrl:
	ld	hl,ctrlcodetable
	ld	b,0
	add	hl,bc
	ld	a,(hl)		; get offset to handler routine
	or	a
	ret	z		; done if no handler
ctrlhandler:
	ld	hl,ctrlhandler	; else get address of handler
	ld	c,a
	add	hl,bc
	push	hl		; put handler address on stack
	ld	hl,(videopos)
	ret			; execute handler
;
bellhandler:
	xor	a
	ld	c,a
bell1:	inc	a
	ld	b,64h
bell2:	out	(soundport),a
	djnz	bell2
	xor	a
	ld	b,64h
bell3:	out	(soundport),a
	djnz	bell3
	dec	c
	jr	nz,bell1
	ret
;
backsphandler:
	ld	a,h
	or	l
	jr	z,incurs1
	dec	h
	jp	p,incurs1
	ld	h,79
	jr	cursorup
;
tabhandler:
	ld	a,h
	and	0f8h
	add	a,8
	ld	h,a
	cp	80
	jr	c,incurs1
	ld	h,0
;
; linefeed handler move cursor down
;
incvidrow:
	inc	l
	jr	incurs1
;
; vtab handler move cursor up
;
cursorup:
	dec	l
	jp	p,incurs1
	ld	l,0
	jr	incurs1
;
; formfeed handler move cursor right
;
inccursor:
	inc	h		; increment column
	ld	a,h
	cp	80
	jr	c,incurs1	; go if not end of row
	ld	h,0		; else set col = 0
	jr	incvidrow	; and increment row
;
; rs handler home cursor
;
homecursor:
	ld	l,0		; set to row 0
crhandler:
	ld	h,0		; set to col 0
incurs1:
	ld	a,l
	cp	24		; at max line?
	ld	(videopos),hl
	ret	c		; done if not at last row
	ld	l,17h		; else set position to bottom row
	ld	(videopos),hl
	ld	hl,videobase+80	; and scroll screen
	ld	de,videobase
	ld	bc,1840
incurs2:
	ldir
incurs3:
	ld	hl,23
	jr	clreos
;
; fs handler insert blank line at cursor
;
insertline:
	call	lengthtoeos
	jr	z,incurs3		; go if at start of last line
	ld	hl,lastvidline-1	; else scroll screen down
	ld	de,-127
	lddr
	ld	hl,(videopos)
	ld	h,0
	jr	clreol
;
; gs handler delete line
;
deleteline:
	call	lengthtoeos
	jr	z,incurs3	; go if at start of last line
	ld	hl,80		; else scroll screen up
	add	hl,de		; add line length to current location
	jr	incurs2
;
; nak handler clear to end of line
;
clreol:	push	hl
	ld	h,50h
	call	getcursadrs
	ex	de,hl
	pop	hl
	jr	clreos1
;
; sub handler clear screen and home cursor
;
cls:	ld	hl,0
	ld	(videopos),hl
;
; etb/em handler clear to end of screen
;
clreos:	ld	de,-128
clreos1:
	call	getcursadrs
	ex	de,hl
	ld	a,(videorevflag)
	or	20h
	sbc	hl,de
	ld	b,h
	ld	c,l
	jp	fillmem
;
; shift out turns reverse video off
;
shiftouthandler:
	xor	a
	jr	setvidflag
;
; shift in turns reverse video on
;
shiftinhandler:
	xor	a
	jr	togvidflag
;
; sync/can toggle reverse video
;
syncanhandler:
	ld	a,(videorevflag)
togvidflag:
	xor	80h
setvidflag:
	ld	(videorevflag),a
	ret
;
eschandler:
	ld	a,3		; start 3 char esc sequence
	jp	escseq6
;
; Get distance from beginning of current screen line
; to beginning of last line into BC
;
lengthtoeos:
	ld	h,0		; set col to 0
	call	getcursadrs
	ex	de,hl
	ld	hl,lastvidline	; address of start of last line
	xor	a
	sbc	hl,de
	ld	b,h		; bc = last line - start of current line
	ld	c,l
	ret
;
cursoronoff:
	ld	a,c
	sub	30h
	jr	z,setcursorflag
	dec	a
	ret	nz
enablecursor:
	ld	a,80h
setcursorflag:
	ld	(cursorflag),a
	ret
;
initvideodata:
	call	clearmem
	dw	videopos
	dw	5
;
	jr	enablecursor
;
; character at current video location
;
videochar:	db	20h
;
; videopos high byte = col, low byte = row
;
videopos:	dw	13h
;
videorevflag:	db	0	; set to 80h when reverse video is enabled
escflag:	db	0	; set when processing esc sequence
cursposrow:	db	11h	; cursor row from escape sequence
;
; If video output character is a control code
; it is translated with this table into an offset
; from the caller to the handler
;
ctrlcodetable:	db	0,0,0,0,0,0,0,0ah
		db	1eh,2ah,36h,39h,41h,4dh,0adh,0b0h
		db	0,0,0,0,0,8ah,0b3h,9ah
		db	8ah,9ah,94h,0bch,6bh,7fh,4bh,0
; 
cursorflag:	db	80h	; cursor on/off flag
;
; Get printer status
; On return A = 0 if printer not ready
;
prtstat:
	in	a,(prtport)
	and	prtstatmask
	xor	30h
	jr	z,prtstat1
	ld	a,1
prtstat1:
	dec	a
	ret
;
; Output character in C to printer
;
prtout:	ld	hl,lptflag
	ld	a,c
	bit	0,(hl)		; suppress lf after cr?
	jr	z,prtot1	; go if not
	cp	lf		; is this a linefeed?
	jr	z,prtot4	; go if so
	cp	cr
	jr	nz,prtot1	; go if not cr
	dec	hl
	ld	a,(hl)		; get last char sent to printer
	inc	hl
	cp	lf
	jr	nz,prtot3	; go if last char wasn't lf
	ld	c,a
	inc	hl		; else increment remaining lines
	inc	(hl)
	dec	hl
prtot1:	cp	formfeed
	jr	nz,prtot3	; go if not formfeed
	bit	1,(hl)		; simulate formfeed?
	jr	z,prtot3	; go if not
	ld	a,(prtremlines)	; else get lines remaining to end of page
	or	a
	jr	z,prtot6	; go if already at end of page
	ld	b,a
prtot2:	call	prtstat		; wait for printer ready
	jr	z,prtot2
	ld	a,lf
	out	(prtport),a	; output lf's until end of page
	djnz	prtot2
	jr	prtot6
;
prtot3:	call	prtstat		; wait for printer ready
	jr	z,prtot3
	ld	a,c
	out	(prtport),a	; print character
prtot4:	ld	a,lf
	cp	c		; was last char a lf?
	ld	a,(prtremlines)	; get remaining lines to end of page
	jr	nz,prtot5	; go if not a lf
	dec	a		; else count down remaining lines
	ld	(prtremlines),a
prtot5:	or	a
	jr	nz,prtot7	; go if not at end of page
prtot6:	ld	a,(prtpagesize)	; else reset line count
	ld	(prtremlines),a
prtot7:	ld	a,c
	ld	(lptchar),a	; save last character printed
	ret
;
initprtflags:
	ld	a,(prtpagesize)
	ld	(prtremlines),a
	xor	a
	ld	(lptchar),a
	ret
;
lptchar:	db	0	; last character sent to printer
;
; Line printer config flag
; bit 0 = suppress linefeed after cr
; bit 1 = simulate formfeed
;
lptflag:	db	1	; config flag
prtremlines:	db	42h	; remaining lines in page
prtpagesize:	db	42h	; default lines per page (66)
;
; TR1602 UART routines
;
constsio:
	in	a,(sioc)	; get uart status
	and	80h		; test data ready bit
	ret	z
	or	0ffh
	ret
;
sioinput:
	in	a,(sioc)	; get uart status
	and	80h		; test data ready bit
	jr	z,sioinput	; loop if no data
	in	a,(siod)	; else read data register
	and	7fh		; mask off bit 7
	ret
;
siooutput:
	call	sioxtst		; test if ready to send
	jr	z,siooutput	; loop if not
	ld	a,c
	out	(siod),a
	ret
;
; test uart xmit ready
;
sioxtst:
	in	a,(sioc)	; get uart status
	and	40h		; test xmit ready bit
	ret	z		; return if not ready
	ld	hl,modemstat
	bit	0,(hl)
	jr	z,sioxt1	; go if cts disabled
	in	a,(modemstatus)
	xor	(hl)
	and	80h		; else test cts bit
	ret	z
sioxt1:	bit	1,(hl)
	jr	z,sioxt2	; go if dsr disabled
	in	a,(modemstatus)
	xor	(hl)
	and	40h		; else test dsr bit
	ret	z
sioxt2:	or	0ffh
	ret
;
setbaudrate:
	ld	a,(defbrgen)
	out	(brgen),a
	out	(rstenable),a
	ld	a,(defsiocsr)
	out	(sioc),a
	ret
;
setbaudvec:
	jp	setbaudrate
;
modemstat:	db	0
defbrgen:	db	0eeh
defsiocsr:	db	6ch
;
; Disk routines
;
; Select disk in C, return dph in HL
;
seldsk:	ld	a,c
	ld	hl,0
	cp	16
	ret	nc		; return if invalid drive
	ld	(seekdisk),a
	ld	(workdisk),a
	ld	b,h
	ld	hl,dph0ptr
	add	hl,bc
	add	hl,bc
	ld	a,(hl)
	inc	hl
	ld	h,(hl)
	ld	l,a
	ld	(seekdphptr),hl
	ret
;
; Set track number to zero
;
home:	ld	bc,0
;
; Set track to number in BC
;
settrk:	ld	(seektrack),bc
	ld	(worktrack),bc
	ret
;
; Set sector to number in BC
;
setsec:	ld	(seeksector),bc
	ld	(worksector),bc
	call	getdpbvec	; get dpb in iy, and dpbvec in ix
	ld	hl,0
	ld	de,(seeksector)
	ld	a,(iy+10h)	; get sector size code from dpb extensions
	or	a
	jr	z,setsec3	; go if 128 byte physical sectors
	ld	b,a
setsec1:
	srl	d		; divide sector by # of 128b blocks in phys sector
	rr	e
	rr	l
	djnz	setsec1
setsec2:
	dec	a		; compute buffer offset from block count
	jr	z,setsec3
	add	hl,hl
	jr	setsec2
;
setsec3:
	ld	(bfroffset),hl
	ld	hl,(seekdphptr)	; get ptr to current dph
	ld	a,(hl)		; get ptr to xlate table
	inc	hl
	ld	h,(hl)		; get xlate table address in hl
	ld	l,a
	or	h
	jr	z,setsec4	; go if no xlate table
	add	hl,de		; else add sector to xlate ptr
	ld	e,(hl)		; and fetch translated sector
setsec4:
	ld	(seeksector),de
	ret
;
; Set dma address in BC
;
setdma:	ld	(dmaptr),bc
	ret
;
; Sector translation in MM is done in setsec.
; Sector passed in BC is returned in HL.
;
sectran:
	ld	h,b
	ld	l,c
	ld	de,0		; why is DE cleared?
	ret
;
; Read sector from disk
;
read:	call	getdpbvec	; get dpb in iy, and dpbvec in ix
	ld	a,1
	call	hostread	; do physical disk access
	ret	nz		; abort read if error occured
	xor	a
	ld	(blockmask),a
;
; Do data transfer for a read or write.
; On entry A = 0 for read, A = 1 for write
;
rdwrxfer:
	ld	hl,(bfroffset)
	ld	de,dskiobfr
	add	hl,de
	ld	de,(dmaptr)
	or	a
	jr	z,rdwrx1	; go if read xfer
	ex	de,hl		; else swap pointers for write xfer
rdwrx1:	call	xferdmabfr	; move data to/from user's buffer
	bit	4,(iy+13h)	; test inverse data bit
	jr	z,rdwrx3	; done if not inverted data
	ld	b,128		; else complement the buffer data
rdwrx2:	dec	de
	ld	a,(de)
	cpl
	ld	(de),a
	djnz	rdwrx2
rdwrx3:	xor	a
	ret
;
xferdmabfr:
	ld	bc,128
	ldir
	ret
;
; Write sector to disk.
; On entry:
; C=0 write can be deferred
; C=1 write must be immediate (directory)
; C=2 write can be deferred, no pre-read is necessary (unallocated space)
;
write:	ld	a,c
	ld	(writecode),a
	call	getdpbvec	; get dpb in iy, and dpbvec in ix
	ld	a,(iy+0fh)	; get physical sectors/track
	cp	6
	jr	z,write2
	ld	a,(writecode)
	cp	2
	jr	nz,write1	; go if immediate write or pre-read necessary
	ld	hl,workdisk
	ld	de,currdisk
	ld	bc,5
	ldir
	ld	l,(iy+0)	; get spt into HL
	ld	h,(iy+1)
	ld	(currspt),hl
	ld	a,(iy+3)	; get BLM
	inc	a
	ld	(blockmask),a
write1:	ld	a,(blockmask)
	or	a
	jr	z,write2
	dec	a
	ld	(blockmask),a
	ld	hl,workdisk
	ld	de,currdisk
	ld	b,5
	call	cmpbytes
	jr	nz,write2	; go if pre-read is required
	ld	hl,currtrk
	call	incsector
	xor	a
	jr	write3		; no pre-read required
;
write2:	xor	a
	ld	(blockmask),a
	inc	a		; pre-read block before write
write3:	call	hostread
	ret	nz		; abort if pre-read failed
	ld	a,1
	call	rdwrxfer	; else transfer write data to i/o buffer
	ld	a,(writecode)
	dec	a
	ld	(writepending),a
	ld	a,0
	ret	nz		; done if deferred write
	jp	flush1		; else flush the write buffer now
;
; On entry HL -> currtrk
;
incsector:
	inc	hl
	inc	hl
	call	inc16hl		; increment currsec
	ld	d,h
	ld	e,l
	inc	de
	inc	de
	call	cmphlde		; compare to maxsec
	ret	nz		; done if not last sector
	inc	hl
	ld	(hl),0		; else zero currsec
	dec	hl
	ld	(hl),0
	dec	hl
	dec	hl
	jp	inc16hl		; and increment track
;
; Get dpb vector for current disk into de & ix
; On return to caller, IY -> dpb
;
getdpbvec:
	ex	(sp),ix		; get return address into ix
	push	iy
	ld	hl,0
	add	hl,sp
	ld	sp,0
	push	hl		; save user sp on system stack
	call	fetchdpbvec
;
; fetchdpbvec will NOT return here, but will
; return to the caller of getdpbvec, and will
; be using the system stack. When the caller of
; the caller of getdpbvec executes a return, it
; will come here, and the user stack will be
; restored along with the user's IX and IY.
;
	pop	hl
	ld	sp,hl		; restore user sp
	pop	iy
	pop	ix
	ret
;
; Get dpb vector for current disk into de & ix
; On exit IY -> dpb, DE & IX -> vector.
; This routine will NOT return to the caller,
; but to the caller of the caller!
;
fetchdpbvec:
	push	ix		; put user return address on system stack
	ld	hl,(seekdphptr)
fetdpbv1:
	ex	de,hl		; de -> current dph
	ld	hl,10
	add	hl,de
	ld	b,(hl)
	inc	hl
	ld	h,(hl)
	ld	l,b
	ex	de,hl		; de -> current dpb
	push	de
	pop	iy		; iy -> current dpb
	ld	e,(iy+11h)	; get dpb vector
	ld	d,(iy+12h)
	push	de		; put dpb vector on stack
fetret:	pop	ix		; ix -> current dpb vector
	ret
;
; Read block from disk (also called for pre-read when writing).
; For read A=1; for write A=1 if pre-read required, else A=0.
;
hostread:
	ld	(preread),a
	ld	hl,seekdisk
	ld	de,hostdisk
	ld	b,7
	call	cmpbytes	; compare 7 bytes at seekdisk & hostdisk
	ret	z		; done if they are the same
	call	flush		; flush any pending write
	jr	nz,fetret	; abort if error
	ld	hl,seekdisk	; transfer disk parameters (disk, track, sector)
	ld	de,hostdisk
	ld	bc,7
	ldir
	ld	a,(hl)		; hl -> preread
	or	a		; need to pre-read for write?
	ret	z		; always fall thru on read
hostrd1:
	ld	a,1		; read function
	call	calldpbvec	; do disk function
	ret	z		; done if no error
	call	diskerr		; else show error message
	jr	hostrd1		; loop if user didn't ^C
;
; Flush any pending write to disk
;
flush:	ld	a,(writepending)
	or	a
	ret	z		; done if no write pending
flush1:	ld	a,2		; else do the write
	call	calldpbvec
	jr	nz,flusherr
	ld	(writepending),a	; clear write pending if no error
	ret
;
flusherr:
	call	diskerr		; show disk error message
	jr	flush1		; loop if user didn't ^C
;
; Called with A=1 for read or A=2 for write
;
calldpbvec:
	push	ix
	push	iy
	ld	hl,(hostdphptr)
	call	fetdpbv1	; get adrs of dpbvector in IX
	ld	bc,(hosttrack)
	ld	de,(hostsector)
	ld	hl,dskiobfr
	call	jumpix		; calls dpbvector to do read or write
	pop	iy
	pop	ix
	ret
;
jumpix:	jp	(ix)
;
; On entry A = fdc error code
;
diskerr:
	and	0fch
	ld	c,a
	ld	a,(lastmemloc)	; get high byte of caller's stack pointer
	cp	bdosbase/256	; high byte of bdos address
	jr	c,chkde3	; ignore if called from below the BDOS
	ld	hl,diskerrcodes
	ld	de,diskerrcode
dsker1:	rl	c
	jr	nc,dsker2
	ld	a,(hl)
	ld	(de),a
	inc	de
dsker2:	inc	hl
	ld	a,c
	or	a
	jr	nz,dsker1
	ld	(de),a
	ld	a,(hostdisk)
	add	a,41h
	ld	(diskltr),a
	ld	hl,diskerrmsg	; display error message
	call	printstr
	call	conin		; get user response
	cp	3
	ret	nz		; if user entered ^C
	jp	0		; reboot!
;
chkde3:	ld	a,1		; ignore error
	pop	hl		; return to caller of caller
	ret
;
diskerrmsg:	db	cr,lf,7,'Disk error on '
diskltr:	db	'x: '
diskerrcode:	db	'      ',0
;
; Disk error codes:
; M = Motor timeout
; W = Write protected disk
; F = Write fault
; S = Sector not found
; C = CRC error
; D = Lost data
;
diskerrcodes:	db	'MWFSCD'
;
; Read/Write routine for ram drive
; Entry with A = 1 read or 2 write
; HL = destination address,
; D = bank to map, E = offset in ram
;
ramrdwr:
	cp	2
	jr	z,ramrw1	; go if write
	call	getbankoffset	; get ram address
	jr	ramrw2		; go do read
;
ramrw1:	call	getbankoffset	; get ram address
	ex	de,hl		; swap pointers for write
ramrw2:	di
	or	8fh
	out	(optreg),a	; map bank to lower ram
	call	xferdmabfr	; transfer 128 bytes
	ld	a,8fh
	out	(optreg),a	; unmap bank
	ei
	xor	a
	ret
;
; Get offset into ram bank
; de <- hl
; hl = e * 128
;
getbankoffset:
	push	hl
	ld	h,e
	ld	l,0
	srl	h
	rr	l
	ld	a,d
	or	6
	rlca
	rlca
	rlca
	rlca
	pop	de
	ret
;
; Read/Write routine for disks A-D
; Entry with A=1 for read or A=2 for write
; HL = dskiobfr, DE = host sector, BC = host track
;
dpbvector:
	dec	a
	jr	z,dpbvrd	; go if read
	dec	a
	jr	z,dpbvwr	; go if write
dpbverr:
	or	80h		; else return error
	ret
;
dpbvrd:	call	seek
	ret	nz		; abort if seek failed
	call	dpbvr2		; try read three times
	ret	z		; done if no error
	ret	m
Xf401:	call	stepinout	; if fail 3 times, move disk head
	call	dpbvr2		; and retry read three more times
	ret	z
	call	restore		; restore head if failed six times
dpbvr2:	call	doread		; try read 3 times
	ret	z		; done if no error
	ret	m
	call	doread
	ret	z
; 
doread:	push	hl		; push anything on stack so nmi will return to rdnmiret
	ld	b,80h		; issue fdc read sector command
	call	dordwr		; set up NMI return to following address
;
	dw	rdnmiret
;
doread1:
	in	a,(fcsr)
	and	e
	jr	z,doread1	; wait for DRQ
	ini			; read first byte from FDC
	ld	a,d
doread2:
	out	(fsel),a
	ini			; keep reading bytes until NMI
	jr	doread2
;
; NMI for read, A = fdc status
;
rdnmiret:
	pop	hl
	and	9ch
	ret
;
dpbvwr:	call	seek
	ret	nz
	call	dpbvw1
	ret	z
	and	0c0h
	ret	nz
	call	stepinout
	call	dpbvw1
	ret	z
	call	restore
dpbvw1:	call	dowrite
	ret	z
	and	0c0h
	ret	nz
	call	dowrite
	ret	z
dowrite:
	push	hl		; push anything on stack so nmi will return to wrnmiret
	ld	b,0a0h		; issue fdc write sector command
	call	dordwr		; set up NMI return to following address
;
	dw	wrnmiret
;
dowrite1:
	in	a,(fcsr)
	and	e
	jr	z,dowrite1	; wait for DRQ
	outi			; write first byte to FDC
	ld	a,(hl)
	inc	hl
	ld	c,fcsr
dowrite2:
	in	e,(c)
	jp	po,dowrite2	; odd parity? this is a goofy way to do things
	out	(fdat),a	; write second byte to FDC
	ld	c,fdat
	ld	a,d
dowrite3:
	out	(fsel),a
	outi			; keep writing until NMI
	jr	dowrite3
;
; NMI for write, A = fdc status
;
wrnmiret:
	pop	hl
	and	0fch
	ret
;
; Seek track routine
;
; On entry HL -> buffer
; DE = host sector, BC = host track
; IY->dpb, IX->dpbvec
; IX + 0,1,2 = vector to rd/wr routine
; IX + 3 = select bit
; IX + 4 = fcsr bits
; IX + 5 = disk motor delay in 1/4 sec (default = 2)
; IX + 6 = head load delay in msec
; IX + 7 = max track + 1
; IX + 8 = write precomp track
; IX + 9 = current track
; IX + 0a = drive select bits
; IX + 0b = seek track
;
seek:	ld	a,(iy+13h)	; get density bit
	and	80h
	or	(ix+3)		; add in drive select bit
	ld	(ix+0bh),c	; set drvsel value (in seek track?)
	bit	6,(iy+13h)	; test density bit
	call	exbios		; exbios just returns, unless exbios.com patched a jump to 0fe80h
	jr	z,seek2		; go if single density
	srl	c		; seek track number /= 2
	bit	2,(iy+13h)	; track numbering same on both sides?
	jr	nz,seek1	; go if not
	ld	(ix+0bh),c	; save seek track number / 2
seek1:	jr	nc,seek2	; go if on side 0
	or	10h		; else set side 1 in drive select bits
	bit	3,(iy+13h)	; sector numbering same for both sides?
	jr	z,seek2		; go if so
	push	af
	ld	a,e		; get sector number
	add	a,(iy+0fh)	; add physical sec/trk to sector for side 1
	ld	e,a
	pop	af
seek2:	ld	(ix+0ah),a	; save updated drive select bits
	ld	a,c		; get track number
	bit	5,(iy+13h)	; test "reserved" bit in options
	jr	z,seek3
	add	a,a		; double track number if "reserved" bit set
seek3:	cp	(ix+8)		; track exceeds write precomp track?
	jr	c,seek4		; go if so
	set	5,(ix+0ah)	; set write precomp bit
seek4:	ld	d,a
	in	a,(fcsr)	; get current fdc status
	rlca
	call	startdrv
	ld	a,0d0h		; issue force interrupt (terminate) command
	out	(fcsr),a
	jr	nc,seek6	; go if fdc was ready
	ld	b,(ix+5)	; get disk motor on delay
seek5:	ld	a,0fah
	call	delayms		; delay 250 msec for each count
	call	startdrv
	djnz	seek5
seek6:	ld	a,e
	out	(fsec),a
seek7:	ld	a,(ix+9)	; set current track in fdc register
	out	(ftrk),a
	inc	a		; if track was -1,
	call	z,doseek	; issue restore command
	ld	a,d		; get seek track
	cp	(ix+7)		; compare to max track + 1
	jp	nc,dpbverr
	out	(fdat),a	; set desired track
	ld	(ix+9),a
	call	startdrv
	in	a,(ftrk)
	sub	d
	jr	z,seek9
	ld	a,d
	or	a
	jr	z,seek8
	ld	a,10h		; fdc seek command
seek8:	call	doseek
seek9:	ld	a,(ix+0bh)
	out	(ftrk),a
	ret
;
restore:
	ld	d,(ix+9)	; get current track
	ld	(ix+9),0ffh	; set track to -1 to force restore
	jr	seek7
;
stepinout:
	ld	a,58h		; stepin command
	call	doseek
	ld	a,68h		; stepout command
doseek:	push	bc
	ld	c,a
	ld	a,2
	call	delayms		; delay 2 msec
	ld	a,(ix+4)	; get step rate from fcsr bits
	and	3
	or	c
	pop	bc
	call	startfdc
doseek1:
	call	startdrv
	in	a,(fcsr)
	rra
	jr	c,doseek1	; wait for BUSY to clear
	ld	a,(ix+6)	; get head load delay
	jp	delayms
;
; Start drive motor
;
startdrv:
	ld	a,(ix+0ah)
	out	(fsel),a
	ret
;
; Issue command to FDC
;
startfdc:
	out	(fcsr),a
	ld	a,14h
strtfdc1:
	dec	a
	jr	nz,strtfdc1
	ret
;
; This routine is called with the stack containing a
; pointer to the nmi return function.
; On entry, B = 80h for read or a0h for write
; On exit, D = drive select bits, E = drq mask, C = fdc data reg
;
dordwr:	push	hl
	ld	a,(nmivec)
	ld	(nmivecsave),a
	ld	hl,(nmiptr)
	ld	(nmiptrsave),hl
	ld	a,jumpcode
	ld	(nmivec),a
	ld	hl,nmisvc
	ld	(nmiptr),hl
	pop	hl
	ex	(sp),hl
	ld	e,(hl)		; get NMI svc return address
	inc	hl
	ld	d,(hl)
	inc	hl
	ex	de,hl
	ex	(sp),hl
	push	de		; put NMI return address on stack
	ld	d,(ix+0ah)	; get drive select bits
	set	6,d		; set bit to generate waits in fsel
	ld	e,2		; set DRQ mask in E
	ld	c,fdat
	call	startdrv
	bit	6,(ix+4)	; test double sided bit
	jr	z,dordwr1	; go if single sided
	set	3,b		; else set side 1 for fdc command
dordwr1:
	di
	ld	a,b
	call	startfdc	; issue fdc command
	ld	a,0c0h
	out	(nmireg),a	; enable NMI
	ret
;
; NMI disk interrupt handler
;
nmisvc:	ex	(sp),hl		; put HL on stack (don't need return adrs)
	xor	a
	out	(nmireg),a	; disable disk interrupts
	ld	a,(nmivecsave)	; restore data at nmi vector
	ld	(nmivec),a
	ld	hl,(nmiptrsave)
	ld	(nmiptr),hl
	in	a,(fcsr)
	pop	hl		; recover HL from stack
	ei
	ret			; return to whoever initiated floppy command
;
; DPH for drive A
;
dph0:	dw	xlt0
	dw	0
	dw	0
	dw	0
	dw	dirbuf
	dw	dpb0
	dw	csv0
	dw	alv0
;
; DPH for drive B
;
dph1:	dw	xlt1
	dw	0
	dw	0
	dw	0
	dw	dirbuf
	dw	dpb1
	dw	csv1
	dw	alv1
;
; DPH for drive C
;
dph2:	dw	xlt2
	dw	0
	dw	0
	dw	0
	dw	dirbuf
	dw	dpb2
	dw	csv2
	dw	alv2
;
; DPH for drive D
;
dph3:	dw	xlt3
	dw	0
	dw	0
	dw	0
	dw	dirbuf
	dw	dpb3
	dw	csv3
	dw	alv3
;
; DPH for ram drive
;
dph4:	dw	0	; no xlate table for ram drive
	dw	0
	dw	0
	dw	0
	dw	dirbuf
	dw	dpb4
	dw	0
	dw	alv4
;
; Each DPB in Montezuma Micro CP/M contains extension
; bytes beyond those specified by CP/M.
;
; Extension bytes are:
; byte 0 = physical sectors/track
; byte 1 = sector size code:
;        0 = 128
;        1 = 256
;        2 = 512
;        3 = 1024
; Documentation implies that bytes 2 and 3 are the number of
; tracks, but actual data is a pointer to the dpb vector.
; byte 2,3 = pointer to dpb vector
; byte 4 = option bits:
;      bit 7 = 0 single density, 1 double density
;      bit 6 = 0 single sided, 1 double sided
;      bit 5 reserved
;      bit 4 = 0 normal data, 1 inverted data
;      bit 3 = 0 sector numbering same on both sides,
;              1 sector numbers on side 1 continue from side 0
;      bit 2 = 0 track number same on both sides,
;              1 track number increments on side change
;      bit 1 = 0 use both sides before changing tracks,
;              1 use all of side 0, then use side 1
;      bit 0 = 0 use side 1 starting at track 0 & go up,
;              1 use side 1 starting at last track & go down
; byte 5 = 0 for ram or hard disk, else is floppy
;              and number is the disk definition entry + 1.
;
; DPB for drive A
;
	if	mmsd40
dpb0:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	0054h	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive A
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb0vec	; dpb2+11h = pointer to dpb vector
	db	80h	; dpb0+13h = option bits
	db	1	; dpb0+14h = drive type code
	endif
;
	if	mmdd40
dpb0:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	00aeh	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive A
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb0vec	; dpb2+11h = pointer to dpb vector
	db	0c0h	; dpb0+13h = option bits
	db	3	; dpb0+14h = drive type code
	endif
;
	if	mmsd80
dpb0:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	00aeh	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive A
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb0vec	; dpb2+11h = pointer to dpb vector
	db	80h	; dpb0+13h = option bits
	db	5	; dpb0+14h = drive type code
	endif
;
	if	mmdd80
dpb0:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	0	; exm
	dw	0162h	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive A
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb0vec	; dpb2+11h = pointer to dpb vector
	db	0c0h	; dpb0+13h = option bits
	db	7	; dpb0+14h = drive type code
	endif
;
	if	mmdde80
dpb0:	dw	28h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	0	; exm
	dw	018ah	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive A
;
	db	0ah	; dpb0+0fh = physical sec/trk
	db	2	; dpb0+10h = sector size code
	dw	dpb0vec	; dpb2+11h = pointer to dpb vector
	db	0c0h	; dpb0+13h = option bits
	db	12	; dpb0+14h = drive type code
	endif
;
; DPB for drive B
;
	if	mmsd40
dpb1:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	0054h	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive B
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb1vec	; dpb2+11h = pointer to dpb vector
	db	80h	; dpb0+13h = option bits
	db	1	; dpb0+14h = drive type code
	endif
;
	if	mmdd40
dpb1:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	00aeh	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive B
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb1vec	; dpb2+11h = pointer to dpb vector
	db	0c0h	; dpb0+13h = option bits
	db	3	; dpb0+14h = drive type code
	endif
;
	if	mmsd80
dpb1:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	00aeh	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive B
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb1vec	; dpb2+11h = pointer to dpb vector
	db	80h	; dpb0+13h = option bits
	db	5	; dpb0+14h = drive type code
	endif
;
	if	mmdd80
dpb1:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	0	; exm
	dw	0162h	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive B
;
	db	12h	; dpb0+0fh = physical sec/trk
	db	1	; dpb0+10h = sector size code
	dw	dpb1vec	; dpb2+11h = pointer to dpb vector
	db	0c0h	; dpb0+13h = option bits
	db	7	; dpb0+14h = drive type code
	endif
;
	if	mmdde80
dpb1:	dw	28h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	0	; exm
	dw	018ah	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive B
;
	db	0ah	; dpb0+0fh = physical sec/trk
	db	2	; dpb0+10h = sector size code
	dw	dpb1vec	; dpb2+11h = pointer to dpb vector
	db	0c0h	; dpb0+13h = option bits
	db	12	; dpb0+14h = drive type code
	endif
;
; DPB for drive C
;
dpb2:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	0aeh	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive C
;
	db	12h	; dpb2+0fh = physical sec/trk
	db	1	; dpb2+10h = sector size code
	dw	dpb2vec	; dpb2+11h = pointer to dpb vector
	db	0c0h	; dpb2+13h = option bits
	db	3	; dpb2+14h = drive type code
;
; DPB for drive D
;
dpb3:	dw	24h	; spt
	db	4	; bsh
	db	0fh	; blm
	db	1	; exm
	dw	0aeh	; dsm
	dw	7fh	; drm
	db	0c0h	; al0
	db	0	; al1
	dw	20h	; cks
	dw	2	; off
;
; MM extensions for drive D
;
	db	12h	; dpb3+0fh = physical sec/trk
	db	1	; dpb3+10h = sector size code
	dw	dpb3vec	; dpb3+11h = pointer to dpb vector
	db	0c0h	; dpb3+13h = option bits
	db	3	; dpb3+14h = drive type code
;
; DPB for ram drive
;
dpb4:	dw	200h	; spt
	db	3	; bsh
	db	7	; blm
	db	0	; exm
	dw	3fh	; dsm
	dw	1fh	; drm
	db	80h	; al0
	db	0	; al1
	dw	0	; cks
	dw	0	; off
;
; MM extensions for ram drive
;
	db	0	; dpb4+0fh = physical sec/trk
	db	0	; dpb4+10h = sector size code
	dw	dpb4vec	; dpb4+11h = pointer to dpb vector
	db	0	; dpb4+13h = option bits
	db	0	; dpb4+14h = drive type code
;
; IX will point here after call to getdpbvec
; IY points to DPB
; Each table is 12 bytes
; IX + 0,1,2 = vector to rd/wr routine
; IX + 3 = drive select bit
; IX + 4 = fcsr bits (accessed by the config utility)
;     bit 7 = number of sides - 1
;     bit 6 = 0 for 5" drive, 1 for 8" drive
;     bits 0-1 = step rate
; IX + 5 = disk motor delay in 1/4 sec (default = 2)
; IX + 6 = head load delay in msec
; IX + 7 = max track + 1
; IX + 8 = write precomp track
; IX + 9 = current track
; IX + 0a = drive select bits
; IX + 0b = seek track
;
dpb0vec:	jp	dpbvector
;
dsk0selbit:	db	1
dsk0fcsr:	db	80h
dsk0motdelay:	db	2
dsk0hdldelay:	db	0fh
	if	mmsd40
dsk0maxtrk:	db	28h
dsk0wptrk:	db	14h
	endif
	if	mmdd40
dsk0maxtrk:	db	28h
dsk0wptrk:	db	14h
	endif
	if	mmsd80
dsk0maxtrk:	db	50h
dsk0wptrk:	db	2ch
	endif
	if	mmdd80
dsk0maxtrk:	db	50h
dsk0wptrk:	db	2ch
	endif
	if	mmdde80
dsk0maxtrk:	db	50h
dsk0wptrk:	db	2ch
	endif
dsk0track:	db	0ffh
dsk0drvsel:	db	81h
dsk0seektrk:	db	0
;
dpb1vec:	jp	dpbvector
;
dsk1selbit:	db	2
dsk1fcsr:	db	80h
dsk1motdelay:	db	2
dsk1hdldelay:	db	0fh
	if	mmsd40
dsk1maxtrk:	db	28h
dsk1wptrk:	db	14h
	endif
	if	mmdd40
dsk1maxtrk:	db	28h
dsk1wptrk:	db	14h
	endif
	if	mmsd80
dsk1maxtrk:	db	50h
dsk1wptrk:	db	2ch
	endif
	if	mmdd80
dsk1maxtrk:	db	50h
dsk1wptrk:	db	2ch
	endif
	if	mmdde80
dsk1maxtrk:	db	50h
dsk1wptrk:	db	2ch
	endif
dsk1track:	db	0ffh
dsk1drvsel:	db	82h
dsk1seektrk:	db	0
;
dpb2vec:	jp	dpbvector
;
dsk2selbit:	db	4
dsk2fcsr:	db	80h
dsk2motdelay:	db	2
dsk2hdldelay:	db	0fh
dsk2maxtrk:	db	28h
dsk2wptrk:	db	14h
dsk2track:	db	0ffh
dsk2drvsel:	db	84h
dsk2seektrk:	db	0
;
dpb3vec:	jp	dpbvector
;
dsk3selbit:	db	8
dsk3fcsr:	db	80h
dsk3motdelay:	db	2
dsk3hdldelay:	db	0fh
dsk3maxtrk:	db	28h
dsk3wptrk:	db	14h
dsk3track:	db	0ffh
dsk3drvsel:	db	88h
dsk3seektrk:	db	0
;
; Vector for ram drive routines
;
dpb4vec:	jp	ramrdwr
;
	if	mmsd40
; Montezuma Micro Standard SYSTEM (40T, SS, DD, 170K)
;
xlt0:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
xlt1:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
	endif
;
	if	mmdd40
; Montezuma Micro Standard DS SYSTEM (40T, DS, DD, 350K)
;
xlt0:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
xlt1:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
	endif
;
	if	mmsd80
; Montezuma Micro 80T SYSTEM (80T, SS, DD, 350K)
;
xlt0:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
xlt1:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
; 
	endif
;
	if	mmdd80
; Montezuma Micro 80T DS SYSTEM (80T, DS, DD, 710K)
;
xlt0:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
xlt1:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
	endif
;
	if	mmdde80
; Montezuma Micro Extended 80T DS SYSTEM (80T, DS, DD, 790K)
;
xlt0:		db	1,3,5,7,9,2,4,6
		db	8,10,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0
;
xlt1:		db	1,3,5,7,9,2,4,6
		db	8,10,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0
;
	endif
;
xlt2:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
xlt3:	db	1,3,5,7,9,11,13,15
	db	17,2,4,6,8,10,12,14
	db	16,18,0,0,0,0,0,0
	db	0,0,0,0,0,0
;
; This is called from the seek routine. If exbios.com was run,
; this will be set to a jp 0fe80h instruction.
;
exbios:	ret
	nop
	nop
;
; The code at 0fe80h will be set by exbios.com to:
; 
; patch:	bit	6,(iy+13h)	; test number of sides
; 		ret	z		; return if single sided
; 		push	de		;
; 		ld	d,a		;
; 		bit	1,(iy+13h)	; do both sides before changing track?
; 		jr	z,patch4	; go if so
; 		ld	e,(ix+7)	;
; 		ld	a,e		;
; 		cp	28h		;
; 		jr	nz,patch1	;
; 		bit	4,(iy+13h)	; inverted data?
; 		jr	z,patch1	; go if not
; 		ld	e,23h		;
; patch1:	ld	a,c		;
; 		sub	e		;
; 		jr	nc,patch2	;
; 		xor	a		;
; 		jr	patch5		;
; ;
; patch2:	bit	0,(iy+13h)	; side 1 start at track 0?
; 		jr	z,patch3	; go if so
; 		neg			;
; 		add	a,e		;
; 		dec	a		;
; patch3:	add	a,a		;
; 		inc	a		;
; 		ld	c,a		;
; patch4:	or	0ffh		;
; patch5:	ld	a,d		;
; 		pop	de		;
; 		ret			;
;
dph0ptr:	dw	dph0
dph1ptr:	dw	dph1
dph2ptr:	dw	dph2
dph3ptr:	dw	dph3
;
		db	0,0,0,0,0,0,0,0
		db	0,0,0,0,0,0,0,0
;
dph4ptr:	dw	0
;
		db	0,0,0,0,0,0
;
seekdisk:	db	0
seekdphptr:	dw	0
seektrack:	dw	0
seeksector:	dw	0
preread:	db	0	; will be 1 if immediate read required
dmaptr:		dw	0
bfroffset:	dw	0
writecode:	db	0
workdisk:	db	0
worktrack:	dw	0
worksector:	dw	0
currdisk:	db	0
currtrk:	dw	0
currsec:	dw	0
currspt:	dw	0
blockmask:	db	0
hostdisk:	db	0
hostdphptr:	dw	0
hosttrack:	dw	0
hostsector:	dw	0
writepending:	db	0
;
	db	0
;
alv0:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
alv1:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
alv2:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
alv3:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
alv4:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
csv0:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
csv1:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
csv2:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
csv3:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
; 
nmivecsave:	db	0
nmiptrsave:	dw	0
;
	end
;

