;
; Model 4 Montezuma Micro CP/M v 2.32
; CCP - Console Command Processor
; for Radio Shack Model 4 computer.
;
; This source code is for the Z80ASM assembler
; from SLR systems.
;
	include	system.inc	; get system definitons
;
biosvec		equ	0
drivecode	equ	4
bdosvec		equ	5
fcb		equ	5ch
dma		equ	80h
tpa		equ	100h
;
ctrlc		equ	3
cr		equ	13
lf		equ	10
ctrlz		equ	1ah
;
DIHALT		equ	76f3h	; Z80 opcodes for DI, HALT
;
	org	ccpbase
;
	jp	command
	jp	clearbfr
;
bfrsize:	db	7fh
bfrcount:	db	0
inputbfr:	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0
	db	'  COPYRIGHT (C) 1979, DIGITAL RE'
	db	'SEARCH  '
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0
	db	0,0
inbfrptr:
	dw	inputbfr
bfrerrptr:
	dw	inputbfr
;
; Print character in A
;
print:	ld	e,a
	ld	c,2
	jp	bdosvec
;
; Print character in A, saving BC
;
printb:	push	bc
	call	print
	pop	bc
	ret
;
; Print CR LF
; 
crlf:	ld	a,cr
	call	printb
	ld	a,lf
	jp	printb
;
; Print SPACE
; 
space:	ld	a,' '
	jp	printb
;
; Print null terminated string pointed to by BC
; 
pline:	push	bc
	call	crlf
	pop	hl
pline1:	ld	a,(hl)
	or	a
	ret	z
	inc	hl
	push	hl
	call	print
	pop	hl
	jp	pline1
;
; Reset disk system
; 
rstdsk:	ld	c,13
	jp	bdosvec
;
; Select disk in A
; 
dsksel:	ld	e,a
	ld	c,14
	jp	bdosvec
;
; Call BDOS and save return code
; 
callbdos:
	call	bdosvec
	ld	(bdosretcode),a
	inc	a		; set Z flag if 0ffh returned
	ret
;
; Open file, DE -> FCB
; 
open:	ld	c,15
	jp	callbdos
;
; Open file at FCB
; 
openfcb:
	xor	a		; clear record number byte
	ld	(ccpfcbrecnum),a	; at fcb + 32
	ld	de,ccpfcb
	jp	open
;
; Close file, DE -> FCB
; 
close:	ld	c,16
	jp	callbdos
;
; Search for first file
; 
srchfst:
	ld	c,17
	jp	callbdos
;
; Search for next file
; 
srchnxt:
	ld	c,18
	jp	callbdos
;
; Search for file at FCB
; 
srchfcb:
	ld	de,ccpfcb
	jp	srchfst
;
; Delete file, DE -> FCB
; 
delete:	ld	c,19
	jp	bdosvec
;
; Call BDOS and test return status
; 
callbdosz:
	call	bdosvec
	or	a
	ret
;
; Read next record, DE -> FCB
; 
readseq:
	ld	c,20
	jp	callbdosz
;
; Read file at FCB
; 
xreadfcb:
	ld	de,ccpfcb
	jp	readseq
;
; Write next record sequential, DE -> FCB
; 
writeseq:
	ld	c,21
	jp	callbdosz
;
; Create file, DE -> FCB
; 
create:	ld	c,22
	jp	callbdos
;
; Rename file, DE -> FCB, new name at DE + 16
; 
renamefile:
	ld	c,23
	jp	bdosvec
;
; Return current user code
; 
getuser:
	ld	e,0ffh
; 
; Get or set user code
; 
getsetuc:
	ld	c,32
	jp	bdosvec
;
; Set current drive in A
; 
setdrive:
	call	getuser		; get user number
	add	a,a		; and shift to upper 4 bits
	add	a,a
	add	a,a
	add	a,a
	ld	hl,crntdrive	; add in drive number
	or	(hl)
	ld	(drivecode),a	; save result
	ret
;
; Set current drive in drive code
; 
fixdrive:
	ld	a,(crntdrive)
	ld	(drivecode),a
	ret
;
; Convert char in A to upper case
; 
toupper:
	cp	'a'
	ret	c
	cp	'z'+1
	ret	nc
	and	5fh
	ret
;
; Get user input line
; If in batch mode, read line from file $$$.sub
; 
getinput:
	ld	a,(batchflag)
	or	a
	jp	z,getinput1	; use console input
	ld	a,(crntdrive)	; select drive 0 for submit file
	or	a
	ld	a,0
	call	nz,dsksel
	ld	de,batchfcb
	call	open		; open $$$.sub file
	jp	z,getinput1	; if missing, use console input
	ld	a,(batchfcblastrec)	; get last record + 1
	dec	a
	ld	(batchfcbrecnum),a
	ld	de,batchfcb
	call	readseq		; read last record
	jp	nz,getinput1	; go if end of file
	ld	de,bfrcount	; else move record into buffer
	ld	hl,dma
	ld	b,128
	call	hl2de		; transfer 128 bytes
	ld	hl,batchfcbs2
	ld	(hl),0		; clear the s2 byte
	inc	hl		; decrement record count
	dec	(hl)
	ld	de,batchfcb	; close the batch file
	call	close
	jp	z,getinput1	; go if error
	ld	a,(crntdrive)	; get previous drive
	or	a
	call	nz,dsksel	; select previous drive if not drive 0
	ld	hl,inputbfr
	call	pline1		; output line to console
	call	chkcon		; check for console input
	jp	z,getinput2	; go if no key pressed
	call	delbatch	; else delete batch file
	jp	command1	; and restart command input
;
getinput1:			; normal console input
	call	delbatch	; delete batch file if it exists
	call	setdrive	; set active drive
	ld	c,10		; get line from console
	ld	de,bfrsize
	call	bdosvec
	call	fixdrive	; reset current drive
getinput2:
	ld	hl,bfrcount	; get character count
	ld	b,(hl)		; and convert line to upper case
getinput3:
	inc	hl
	ld	a,b
	or	a
	jp	z,getinput4	; go if end of line
	ld	a,(hl)
	call	toupper
	ld	(hl),a
	dec	b
	jp	getinput3
;
getinput4:
	ld	(hl),a		; null terminate input line
	ld	hl,inputbfr	; reset input line pointer
	ld	(inbfrptr),hl
	ret
;
; Check console for keypress
; A = character or 0 for none
; 
chkcon:	ld	c,11
	call	bdosvec
	or	a
	ret	z		; done if no input
	ld	c,1
	call	bdosvec		; else read the character
	or	a
	ret
;
; Get current drive number
; 
getdsk:	ld	c,25
	jp	bdosvec
;
; Set standard dma address
; 
stddma:	ld	de,dma
;
; Set dma address to DE
;
dmaset:	ld	c,26
	jp	bdosvec
;
; Delete batch file created by SUBMIT
; 
delbatch:
	ld	hl,batchflag	; check if batch mode
	ld	a,(hl)		; and return if not
	or	a
	ret	z
	ld	(hl),0		; clear batch mode
	xor	a
	call	dsksel		; select drive 0
	ld	de,batchfcb	; delete $$$.sub file
	call	delete
	ld	a,(crntdrive)	; reset current drive
	jp	dsksel
;
; Check serial number in BDOS
; If serial numbers don't match, halt the machine
; 
checkserno:
	ld	de,serialnum
	ld	hl,bdosbase
	ld	b,6
chkserno1:
	ld	a,(de)
	cp	(hl)
	jp	nz,crash	; mismatch, go die
	inc	de
	inc	hl
	dec	b
	jp	nz,chkserno1
	ret
;
; Print user's command with '?'
; 
syntaxerr:
	call	crlf
	ld	hl,(bfrerrptr)	; points to command
syntx1:	ld	a,(hl)		; print til space or null
	cp	' '
	jp	z,syntx2
	or	a
	jp	z,syntx2
	push	hl
	call	print
	pop	hl
	inc	hl
	jp	syntx1
;
syntx2:	ld	a,'?'		; add '?' at end
	call	print
	call	crlf
	call	delbatch	; delete batch file if any
	jp	command1	; and redo command input
;
; Check char at (DE) for command input
; Set Z if char is a delimiter ( = _ . : ; < > )
; 
check:	ld	a,(de)
	or	a
	ret	z
	cp	' '
	jp	c,syntaxerr	; abort if char < space
	ret	z
	cp	'='
	ret	z
	cp	'_'
	ret	z
	cp	'.'
	ret	z
	cp	':'
	ret	z
	cp	';'
	ret	z
	cp	'<'
	ret	z
	cp	'>'
	ret	z	; unnecessary, but that's the way Gary wrote it
	ret
;
; Get the next non-blank char from (DE)
; 
nonblank:
	ld	a,(de)
	or	a
	ret	z		; done if null character
	cp	' '
	ret	nz		; done if not space
	inc	de
	jp	nonblank
;
; Add HL = HL + A
; 
addhl:	add	a,l
	ld	l,a
	ret	nc
	inc	h		; if carry, add to high byte
	ret
;
; Convert first file name in FCB
; 
convfst:
	ld	a,0
; 
; Format a file name (convert * to ? etc)
; Return A=0 if unambiguous file name
; On entry A = FCB position (0 or 16)
; 
convert:
	ld	hl,ccpfcb
	call	addhl
	push	hl
	push	hl
	xor	a
	ld	(drvchngflag),a	; init drive change flag
	ld	hl,(inbfrptr)	; set input line pointer
	ex	de,hl		; into DE
	call	nonblank	; get next non-blank character
	ex	de,hl
	ld	(bfrerrptr),hl	; save pointer in case of error
	ex	de,hl
	pop	hl
	ld	a,(de)		; get first character
	or	a
	jp	z,convrt1
	sbc	a,'A'-1		; convert to binary in case it's a drive name
	ld	b,a
	inc	de
	ld	a,(de)		; check next character for drive designation
	cp	':'
	jp	z,convrt2	; go if drive name
	dec	de		; else point back to start of input
convrt1:
	ld	a,(crntdrive)
	ld	(hl),a
	jp	convrt3
;
convrt2:
	ld	a,b		; drive number in B
	ld	(drvchngflag),a	; save it
	ld	(hl),b
	inc	de
;
; now convert the file name
;
convrt3:
	ld	b,8
convrt4:
	call	check		; check for delimiter ( = _ . : ; < > )
	jp	z,convrt8
	inc	hl
	cp	'*'
	jp	nz,convrt5	; if '*', fill remainder with '?'
	ld	(hl),'?'
	jp	convrt6
;
convrt5:
	ld	(hl),a
	inc	de
convrt6:
	dec	b
	jp	nz,convrt4
convrt7:
	call	check		; check for delimiter
	jp	z,getext
	inc	de
	jp	convrt7
;
convrt8:
	inc	hl
	ld	(hl),' '	; blank fill remainder of file name
	dec	b
	jp	nz,convrt8
;
; now convert the file extension
;
getext:	ld	b,3
	cp	'.'
	jp	nz,getext5	; skip over '.' separator
	inc	de
getext1:
	call	check		; check for delimiter
	jp	z,getext5
	inc	hl
	cp	'*'
	jp	nz,getext2
	ld	(hl),'?'	; if wildcard, fill with '?'
	jp	getext3
;
getext2:
	ld	(hl),a
	inc	de
getext3:
	dec	b
	jp	nz,getext1
getext4:
	call	check
	jp	z,getext6
	inc	de
	jp	getext4
;
getext5:
	inc	hl
	ld	(hl),' '	; blank fill remainder of extension
	dec	b
	jp	nz,getext5
getext6:
	ld	b,3
getext7:
	inc	hl
	ld	(hl),0
	dec	b
	jp	nz,getext7
	ex	de,hl
	ld	(inbfrptr),hl
	pop	hl
	ld	bc,11		; check file name and extension
getext8:
	inc	hl
	ld	a,(hl)
	cp	'?'
	jp	nz,getext9
	inc	b		; count question marks
getext9:
	dec	c
	jp	nz,getext8
	ld	a,b
	or	a		; A != 0 if ambiguous file spec
	ret
;
cmdtable:
	db	'DIR '
	db	'ERA '
	db	'TYPE'
	db	'SAVE'
	db	'REN '
	db	'USER'
;
serialnum:	db	0,0,0,0,0,0
; 
; Check for built-in command.
; Return command number in A (A = 6 if not built-in command).
;
search:	ld	hl,cmdtable
	ld	c,0
search1:
	ld	a,c
	cp	6
	ret	nc		; done, no match
	ld	de,ccpfcbfname	; command or file name from fcb
	ld	b,4		; set command length
search2:
	ld	a,(de)		; get next character of command
	cp	(hl)
	jp	nz,search3	; doesn't match current cmdtable entry
	inc	de
	inc	hl
	dec	b
	jp	nz,search2	; check all characters
	ld	a,(de)		; if four character command,
	cp	' '		; next char in fcb must be a space, else it might
	jp	nz,search4	; be a file name that starts with a command name
	ld	a,c		; return command number in A
	ret
;
search3:
	inc	hl		; point to next entry in cmdtable
	dec	b
	jp	nz,search3
search4:
	inc	c		; increment command number
	jp	search1		; and check next entry
;
; Clear the command input buffer and start CCP
;
clearbfr:
	xor	a
	ld	(bfrcount),a
;
; Start console command processor
; C = user number in high 4 bits, drive number in low 4 bits
;
command:
	ld	sp,batchflag	; set ccp stack
	push	bc
	ld	a,c		; get user number from C
	rra
	rra
	rra
	rra
	and	0fh
	ld	e,a
	call	getsetuc	; set user number
	call	rstdsk		; reset disk system
	ld	(batchflag),a	; clear batch mode
	pop	bc
	ld	a,c
	and	0fh		; get drive number from C
	ld	(crntdrive),a	; save it,
	call	dsksel		; and select the drive
	ld	a,(bfrcount)	; check input buffer
	or	a
	jp	nz,command2	; process input if any
;
; Get input line from console
;
command1:
	ld	sp,batchflag	; set ccp stack
	call	crlf		; begin new line on console
	call	getdsk		; get current drive
	add	a,'A'		; display on console
	call	print
	ld	a,'>'		; display prompt
	call	print
	call	getinput	; get input line from user or batch file
command2:
	ld	de,dma		; set default dma address
	call	dmaset
	call	getdsk
	ld	(crntdrive),a	; set current drive
	call	convfst		; expand wildcards in input
	call	nz,syntaxerr	; abort if syntax error
	ld	a,(drvchngflag)
	or	a		; if change in drive, then assume
	jp	nz,load		; it's an external command
	call	search		; search for command in cmdtable
	ld	hl,cmdadrstbl	; A = command number
	ld	e,a		; get command address from table
	ld	d,0
	add	hl,de
	add	hl,de
	ld	a,(hl)		; load address of routine into HL
	inc	hl
	ld	h,(hl)
	ld	l,a
	jp	(hl)		; execute the built in command
;
cmdadrstbl:
	dw	direct
	dw	erase
	dw	type
	dw	save
	dw	rename
	dw	user
	dw	load		; unknown commands are external
;
; Halt the system
; This gets invoked if the serial number in CCP doesn't match
; the serial number in BDOS. (Paranoia from DR)
; 
crash:	ld	hl,DIHALT	; set di, halt commands at start of ccp
	ld	(ccpbase),hl
	ld	hl,ccpbase	; then go execute the halt code
	jp	(hl)
;
; File read error with TYPE command
;
readerr:
	ld	bc,rderrmsg
	jp	pline
;
rderrmsg:	db	'READ ERROR',0
;
; File could not be found
;
nofile:	ld	bc,nofilemsg
	jp	pline
;
nofilemsg:	db	'NO FILE',0
; 
; Decode a 'save' or 'user' command followed by a number
;
decode:	call	convfst		; expand wild cards in filename
	ld	a,(drvchngflag)
	or	a
	jp	nz,syntaxerr	; drive change not allowed in these commands
	ld	hl,ccpfcbfname
	ld	bc,11		; init result (B) and counter (C)
decode1:
	ld	a,(hl)
	cp	' '
	jp	z,decode2	; go if end of number
	inc	hl
	sub	'0'		; else check for valid digit
	cp	10
	jp	nc,syntaxerr	; error if digit > 9
	ld	d,a		; else save digit
	ld	a,b
	and	0e0h		; limit number to 0-31
	jp	nz,syntaxerr	; go if result exceeds limit
	ld	a,b		; else multiply result by 10
	rlca			; * 2
	rlca			; * 4
	rlca			; * 8
	add	a,b		; + 1 = * 9
	jp	c,syntaxerr	; error if overflow
	add	a,b		; + 1 = * 10
	jp	c,syntaxerr	; error if overflow
	add	a,d		; add in new digit
	jp	c,syntaxerr	; error if overflow
	ld	b,a		; save new result
	dec	c
	jp	nz,decode1	; do for max of 11 digits
	ret
;
decode2:
	ld	a,(hl)		; remainder of command must be spaces
	cp	' '		; since it comes from a 'filename'
	jp	nz,syntaxerr	; error if not a space
	inc	hl
	dec	c
	jp	nz,decode2
	ld	a,b		; return result in A
	ret
;
; Move 3 bytes from (HL) to (DE)
;
move3bytes:
	ld	b,3
;
; Move B bytes from (HL) to (DE)
; Like LDIR, but only uses B as a counter
; 
hl2de:	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	dec	b
	jp	nz,hl2de
	ret
;
; Compute HL = dma + A + C and get the byte there
;
extract:
	ld	hl,dma
	add	a,c
	call	addhl
	ld	a,(hl)
	ret
;
; Select the drive specified in drvchngflag
;
dselect:
	xor	a		; clear first byte of fcb to
	ld	(ccpfcb),a	; indicate current drive
	ld	a,(drvchngflag)	; and if drive change is required,
	or	a
	ret	z
	dec	a
	ld	hl,crntdrive
	cp	(hl)		; is new drive same as current drive?
	ret	z		; no need to select if so
	jp	dsksel		; select the new drive
;
; If last command required a drive change,
; reselect the previous drive
;
resetdr:
	ld	a,(drvchngflag)	; was there a drive change?
	or	a
	ret	z		; done if not
	dec	a
	ld	hl,crntdrive	; was it same as current drive?
	cp	(hl)
	ret	z		; done if so
	ld	a,(crntdrive)	; else reselect current drive
	jp	dsksel
;
; Directory command (DIR)
; 
direct:	call	convfst		; expand wildcards in filename
	call	dselect		; select the drive
	ld	hl,ccpfcbfname
	ld	a,(hl)
	cp	' '
	jp	nz,dir2		; go if file name supplied
	ld	b,11
dir1:	ld	(hl),'?'	; else fill filename with '?'
	inc	hl
	dec	b
	jp	nz,dir1
dir2:	ld	e,0
	push	de
	call	srchfcb		; find first file match
	call	z,nofile	; go if no match
dir3:	jp	z,dir11		; done if no more file matches
	ld	a,(bdosretcode)	; get file's fcb position
	rrca
	rrca
	rrca
	and	60h
	ld	c,a
	ld	a,10
	call	extract		; fetch byte in position 10
	rla			; check file system status bit
	jp	c,dir10		; don't list if system file
	pop	de
	ld	a,e
	inc	e		; increment file count
	push	de
	and	3
	push	af
	jp	nz,dir4		; go if not fourth file on a line
	call	crlf		; else start a new line
	push	bc
	call	getdsk		; get disk number
	pop	bc
	add	a,'A'		; output drive name
	call	printb
	ld	a,':'
	call	printb
	jp	dir5
;
dir4:	call	space		; separate file names
	ld	a,':'
	call	printb
dir5:	call	space
	ld	b,1
dir6:	ld	a,b
	call	extract		; get next character in file name
	and	7fh		; remove status bit
	cp	' '
	jp	nz,dir8		; go if not end of name
	pop	af
	push	af
	cp	3
	jp	nz,dir7
	ld	a,9
	call	extract		; check for file extension
	and	7fh		; remove status bit
	cp	' '
	jp	z,dir9		; go if no extension
dir7:	ld	a,' '		; else print filename spaces
dir8:	call	printb
	inc	b		; next character position
	ld	a,b
	cp	12
	jp	nc,dir9		; go if end of name
	cp	9
	jp	nz,dir6		; go if not start of extension
	call	space		; else output a space
	jp	dir6
;
dir9:	pop	af
dir10:	call	chkcon		; check console input
	jp	nz,dir11	; abort if any input
	call	srchnxt		; get next matching filename
	jp	dir3
;
dir11:	pop	de
	jp	getback
;
; Erase command
; 
erase:	call	convfst		; expand wildcards in filename
	cp	11
	jp	nz,erase1	; go if not '*.*'
	ld	bc,allprompt	; else ask to confirm
	call	pline
	call	getinput
	ld	hl,bfrcount	; response must be one character
	dec	(hl)
	jp	nz,command1	; abort if not one character
	inc	hl
	ld	a,(hl)
	cp	'Y'
	jp	nz,command1	; abort if not 'Y'
	inc	hl
	ld	(inbfrptr),hl	; save input pointer
erase1:	call	dselect		; select the disk
	ld	de,ccpfcb
	call	delete		; delete the file
	inc	a
	call	z,nofile	; call if file not found
	jp	getback
;
allprompt:	db	'ALL (Y/N)?',0
; 
; Type command
;
type:	call	convfst		; expand wildcards in filename
	jp	nz,syntaxerr	; wildcards not allowed for TYPE
	call	dselect		; select the indicated drive
	call	openfcb		; open the file
	jp	z,type4		; go if file missing
	call	crlf		; else output a new line
	ld	hl,numbytes
	ld	(hl),0ffh	; init byte counter
type1:	ld	hl,numbytes
	ld	a,(hl)
	cp	128
	jp	c,type2		; go if not done with this sector
	push	hl
	call	xreadfcb	; read the next sector
	pop	hl
	jp	nz,type3	; end of file or read error
	xor	a
	ld	(hl),a		; clear byte counter
type2:	inc	(hl)		; count bytes output
	ld	hl,dma		; get next byte from buffer
	call	addhl
	ld	a,(hl)
	cp	ctrlz		; end of file?
	jp	z,getback	; done if so
	call	print		; else output character
	call	chkcon		; check console input
	jp	nz,getback	; abort if any input
	jp	type1		; else do next character
;
type3:	dec	a		; read error?
	jp	z,getback	; if not, end of file and we're done
	call	readerr		; else output error message
type4:	call	resetdr		; reset the drive
	jp	syntaxerr	; and output file name
;
; Save command
; 
save:	call	decode		; get number after save command
	push	af		; save pages to write
	call	convfst		; convert file name
	jp	nz,syntaxerr	; wild cards not allowed
	call	dselect		; select the drive
	ld	de,ccpfcb	; and delete the file if it exists
	push	de
	call	delete
	pop	de
	call	create		; make a new file
	jp	z,save3		; abort if can't create
	xor	a
	ld	(ccpfcbrecnum),a
	pop	af		; convert pages to records
	ld	l,a
	ld	h,0
	add	hl,hl		; number of records to write into HL
	ld	de,tpa
save1:	ld	a,h		; test number of records
	or	l
	jp	z,save2		; done if no more left
	dec	hl
	push	hl
	ld	hl,128		; get address of next record
	add	hl,de
	push	hl
	call	dmaset		; set dma to record address
	ld	de,ccpfcb
	call	writeseq	; write record to file
	pop	de
	pop	hl
	jp	nz,save3	; go if disk full
	jp	save1
;
save2:	ld	de,ccpfcb	; done, close the file
	call	close
	inc	a
	jp	nz,save4	; go if close ok
save3:	ld	bc,dskfullmsg
	call	pline
save4:	call	stddma
	jp	getback
;
dskfullmsg:	db	'NO SPACE',0
; 
; Rename command
;
rename:	call	convfst		; convert first file name
	jp	nz,syntaxerr	; wild cards not allowed
	ld	a,(drvchngflag)
	push	af
	call	dselect		; select specified drive
	call	srchfcb		; does the file exist?
	jp	nz,rename5	; abort if so
	ld	hl,ccpfcb	; else save the file name
	ld	de,ccpfcbfile2
	ld	b,16
	call	hl2de
	ld	hl,(inbfrptr)	; get input buffer pointer
	ex	de,hl
	call	nonblank	; find second file name
	cp	'='
	jp	z,rename1	; '=' separator allowed
	cp	'_'
	jp	nz,rename4	; underscore separator allowed
rename1:
	ex	de,hl
	inc	hl		; skip over separator
	ld	(inbfrptr),hl	; save input pointer
	call	convfst		; convert second file name
	jp	nz,rename4	; wild cards not allowed
	pop	af
	ld	b,a
	ld	hl,drvchngflag
	ld	a,(hl)
	or	a
	jp	z,rename2	; if drive specified,
	cp	b		; it must be the same as destination drive
	ld	(hl),b
	jp	nz,rename4	; error if drives not the same
rename2:
	ld	(hl),b
	xor	a
	ld	(ccpfcb),a	; clear drive byte in fcb
	call	srchfcb		; does source file exist?
	jp	z,rename3	; error if not
	ld	de,ccpfcb
	call	renamefile	; else rename the file
	jp	getback
;
rename3:
	call	nofile		; source file doesn't exist
	jp	getback
;
rename4:
	call	resetdr		; source and destination drives not the same
	jp	syntaxerr
;
rename5:
	ld	bc,filexistmsg	; destination file exists
	call	pline
	jp	getback
;
filexistmsg:	db	'FILE EXISTS',0
; 
; User command
;
user:	call	decode		; get user number
	cp	16
	jp	nc,syntaxerr	; abort if not 0-15
	ld	e,a		; save user number
	ld	a,(ccpfcbfname)
	cp	' '		; anything following user number?
	jp	z,syntaxerr	; error if so
	call	getsetuc	; else set user number
	jp	getback1
;
; Load transient command from disk
; 
load:	call	checkserno	; check serial number (paranoia)
	ld	a,(ccpfcbfname)	; command file specified?
	cp	' '
	jp	nz,load1	; go if so
	ld	a,(drvchngflag)	; drive change requested?
	or	a
	jp	z,getback1	; go if not
	dec	a
	ld	(crntdrive),a	; else save new drive number
	call	fixdrive
	call	dsksel		; select the drive
	jp	getback1	; and we're done
;
; command file name was entered
;
load1:	ld	de,ccpfcbext	; was there an extension?
	ld	a,(de)
	cp	' '
	jp	nz,syntaxerr	; error if file extension entered
	push	de
	call	dselect		; select the specified drive
	pop	de
	ld	hl,commtext	; set extension to 'COM'
	call	move3bytes
	call	openfcb		; open the file
	jp	z,loaderr	; abort if file not found
	ld	hl,tpa		; else init the transfer address
load2:	push	hl
	ex	de,hl
	call	dmaset		; set transfer address
	ld	de,ccpfcb
	call	readseq		; read next record
	jp	nz,load3	; go if end of file or read error
	pop	hl
	ld	de,128
	add	hl,de		; update memory pointer
	ld	de,ccpbase	; test if overwrite of CCP
	ld	a,l
	sub	e
	ld	a,h
	sbc	a,d
	jp	nc,loaderr1	; go if it would overwrite CCP
	jp	load2
;
; Done reading command file
;
load3:	pop	hl
	dec	a		; check for end of file
	jp	nz,loaderr1	; go if read error
	call	resetdr		; else reset to previous drive
	call	convfst
	ld	hl,drvchngflag
	push	hl
	ld	a,(hl)
	ld	(ccpfcb),a
	ld	a,16
	call	convert		; convert second file name (if any)
	pop	hl
	ld	a,(hl)
	ld	(ccpfcbfile2),a	; set drive code for second file
	xor	a
	ld	(ccpfcbrecnum),a	; clear record number
	ld	de,fcb
	ld	hl,ccpfcb	; set up second fcb
	ld	b,33
	call	hl2de
	ld	hl,inputbfr	; move remainder of input line to dma buffer
load4:	ld	a,(hl)
	or	a
	jp	z,load5		; search for null
	cp	' '
	jp	z,load5		; or space
	inc	hl
	jp	load4
;
load5:	ld	b,0		; init character count
	ld	de,dma+1
load6:	ld	a,(hl)		; move command line remainder to dma buffer
	ld	(de),a
	or	a
	jp	z,load7		; done if null terminator
	inc	b
	inc	hl
	inc	de
	jp	load6
;
load7:	ld	a,b
	ld	(dma),a		; save the character count in dma buffer
	call	crlf		; output a new line
	call	stddma		; set standard dma address
	call	setdrive	; select the current drive
	call	tpa		; and execute the program
;
; transient programs either reboot or return here
;
	ld	sp,batchflag	; reset ccp stack
	call	fixdrive
	call	dsksel		; select current drive
	jp	command1	; and go back to command input
;
loaderr:
	call	resetdr
	jp	syntaxerr
;
loaderr1:
	ld	bc,badloadmsg
	call	pline
	jp	getback
;
badloadmsg:	db	'BAD LOAD',0
commtext:	db	'COM'
; 
; Return to command level
;
getback:
	call	resetdr		; reset previous drive
getback1:
	call	convfst		; convert first filename if fcb
	ld	a,(ccpfcbfname)
	sub	' '
	ld	hl,drvchngflag	; if just a drive change,
	or	(hl)		; make sure it was a valid drive
	jp	nz,syntaxerr	; go if not
	jp	command1	; else return to command mode
;
; CCP stack area, initialized to reboot in bios
;
stackend:
	dw	biosvec
	dw	biosvec
	dw	biosvec
	dw	biosvec
	dw	biosvec
	dw	biosvec
	dw	biosvec
	dw	biosvec
;
; ccp stack starts here
; 
batchflag:	db	0
batchfcb:	db	0
subtext:	db	'$$$     SUB',0,0
batchfcbs2:	db	0
batchfcblastrec:
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0,0,0,0,0
batchfcbrecnum:	db	0
; 
ccpfcb:	db	0
ccpfcbfname:	db	'        '
ccpfcbext:	db	'   '
	db	0,0,0,2bh
ccpfcbfile2:	db	0,20h,20h,20h,20h,20h,20h,20h
	db	20h,20h,20h,20h,0,0,0,0
ccpfcbrecnum:	db	0
bdosretcode:	db	1
crntdrive:	db	0
drvchngflag:	db	0
numbytes:
	dw	4eh
;
	db	0,0,0,0,0,0,0,0
	db	0,0,0,0,0
;
	end
;

