;
; Montezuma Micro CP/M 2.32
; for Radio Shack Model 4 computer.
;
; CP/M utility to combine assembled BOOT, CCP, BDOS, and BIOS
; into a SYSTEM file for use with modified SYSGEN utility.
;
; Usage:
; "combine bootxx.com ccpxx.com bdosxx.com biosxx.com sysxx.bin"
; where xx is the system size (ie: sys63.bin is a 63K system).
; The last command line argument is the output file.
;
; This source code is for the Z80ASM assembler
; from SLR systems.
;
; The boot sector code is 256 bytes loaded by the Model 4 rom.
; CCP is 2,048 (0800h) bytes.
; BDOS is 3,584 (0e00h) bytes.
; BIOS is 3,840 (0f00h) bytes.
; Total size of combined file for use by sysgen is 9,728 (2600h) bytes.
;
bdos	equ	5
fcb	equ	5ch
fcb2	equ	6ch
dma	equ	80h
tpa	equ	100h
fcbsize	equ	36
;
cr	equ	13
lf	equ	10
;
read	equ	0
write	equ	1
;
	org	tpa
;
	jp	start
;
argc:	db	0		; argument count (does not include name of program)
argv:	defs	32*2		; argument vector table
cmdlin:	defs	80		; command line buffer
;
sysstk:	defw	0		; system stack pointer
flen:	defw	0		; length of output file (sum of input files)
bfrptr:	defw	buffer		; current pointer into buffer
;
outfname:	defs	16	; output file name
;
outfcb:	defs	fcbsize
infcb:	defs	fcbsize
;
start:	call	parse		; parse cmd line args
	ld	a,(argc)
	or	a
	jr	nz,main
	ld	hl,noargmsg
	call	print
	ret
;
; The main program.
;
main:	ld	(sysstk),sp
	ld	sp,stack
	ld	hl,argv		; get output file name (last argv)
	ld	a,(argc)
	dec	a
	add	a,a
	ld	b,0
	ld	c,a
	add	hl,bc
	ld	a,(hl)
	inc	hl
	ld	h,(hl)
	ld	l,a
	ld	de,outfname	; set up output file name
main1:	ld	a,(hl)
	ld	(de),a
	inc	hl
	inc	de
	or	a
	jr	nz,main1
	ld	hl,outfname
	ld	de,infcb
	call	cpyfnm
	ld	a,(argc)	; get input file count
	dec	a
	ld	b,a
	ld	c,0		; init file count
;
; Main loop. Read each input file into buffer and
; when all files read, write buffer to output file.
;
main2:	ld	hl,argv
	ld	a,c		; current file number
	add	a,a
	ld	e,a
	ld	d,0
	add	hl,de
	ld	a,(hl)
	inc	hl
	ld	h,(hl)
	ld	l,a
	or	a
	jr	z,exit
	call	rdfile		; read file into buffer
	ld	a,h
	or	l
	jr	z,rderror
	ld	de,(flen)
	add	hl,de
	ld	(flen),hl	; save file size
	inc	c		; next input file
	djnz	main2
	ld	hl,outfname
	ld	hl,buffer
	ld	(bfrptr),hl
	ld	hl,outfname
	call	wrfile		; write buffer to output file
	call	crlf
;
exit:	ld	sp,(sysstk)	; return to CP/M
	ret
;
rderror:
	ld	hl,rderrmsg
	call	print
	jr	exit
;
; fopen(char *fname, FCB *fcb, int rdwr);
; On entry, HL points to filename, DE points to FCB.
;           A = 0 if read, 1 if write.
; Returns NZ if open, Z if failure
;
fopen:	push	hl
	push	de
	push	bc
	push	de		; save fcb
	push	af		; save r/w code
	ld	b,fcbsize	; clear fcb first
	xor	a
fopen1:	ld	(de),a
	inc	de
	djnz	fopen1
	pop	af		; restore r/w code
	pop	de		; restore fcb
	push	af
	push	de		; save fcb
	call	cpyfnm		; copy filename to fcb
	pop	de		; restore fcb
	pop	af
	or	a
	jr	z,fopenr	; go if reading
	ld	c,17		; else see if output file exists
	push	de		; save fcb
	call	bdos
	pop	de		; restore fcb
	inc	a
	jr	z,fopenc	; file does not exist, so create it
	push	de		; else delete it (save fcb)
	ld	c,19
	call	bdos
	pop	de		; restore fcb
fopenc:	push	de		; save fcb
	ld	c,22
	call	bdos		; and create new file
	pop	de		; restore fcb
	inc	a
	jr	z,fopenx	; go if file not created
fopenr:	ld	c,15
	call	bdos		; open file
	inc	a
fopenx:	pop	bc
	pop	de
	pop	hl
	ret
;
; Copy file name pointed to by HL into fcb pointed to by DE.
; File name may be "file.bin", and should be expanded to
; "file   bin" in fcb.
;
cpyfnm:	inc	de		; start of file name in fcb
	ld	a,(hl)
	cp	' '
	jr	nz,cpfnm0	; skip leading space, if any
	inc	hl
cpfnm0:	ld	b,8		; size of fcb file name
cpfnm1:	ld	a,(hl)		; set up file name
	inc	hl
	cp	'.'
	jr	z,cpfnm2	; go if extension separator
cpf1a:	ld	(de),a		; else save in fcb
	inc	de		; next character
	djnz	cpfnm1
cpfnm2:	ld	a,b		; was file name full 8 characters?
	or	a
	jr	z,cpfnm4	; go if so
	ld	a,' '		; else fill remainder with spaces
cpfnm3:	ld	(de),a
	inc	de
	djnz	cpfnm3
cpfnm4:	ld	b,3		; size of fcb extension
	ld	a,(hl)
	cp	'.'
	jr	nz,cpfnm5
	inc	hl
cpfnm5:	ld	a,(hl)
	inc	hl
	or	a
	jr	z,cpfnm6	; go if end of extension
	ld	(de),a
	inc	de
	djnz	cpfnm5
cpfnm6:	ld	a,b
	or	a		; was extension full 3 characters?
	jr	z,cpfnm8		; done if so
	ld	a,' '		; else fill remainder with spaces
cpfnm7:	ld	(de),a
	inc	de
	djnz	cpfnm7
cpfnm8:	xor	a
	ld	(de),a
	ret
;
; fclose();
; On entry, DE points to fcb
; Returns NZ if open, Z if failure.
;
fclose:	push	hl
	push	de
	push	bc
	ld	c,16
	call	bdos		; close file
	inc	a
	pop	bc
	pop	de
	pop	hl
	ret
;
; Read next file into buffer.
; On entry, HL points to file name.
; On return, HL = number of bytes read.
;
rdfile:	push	de
	push	bc
	push	hl
	ld	hl,rdmsg
	call	print
	pop	hl
	push	hl
	call	print
	pop	hl
	ld	a,read
	ld	de,infcb
	call	fopen
	jr	nz,rdfil1
	ld	hl,fopenerrmsg
	call	print
	ld	hl,0
	jr	rdfilx
;
rdfil1:	ld	hl,0		; init byte cout
rdfil2:	push	hl
	ld	c,20		; read next block
	ld	de,infcb
	call	bdos
	pop	hl
	or	a		; successful read?
	jr	nz,rdfild	; go if done
	ld	bc,128
	add	hl,bc
	push	hl
	ld 	hl,(bfrptr)	; transfer data to buffer
	ex	de,hl
	ld	hl,dma
	ldir
	ex	de,hl
	ld	(bfrptr),hl
	pop	hl
	jr	rdfil2		; get next block
;
rdfild:	ld	a,' '
	call	chrout
	ld	a,'-'
	call	chrout
	ld	a,' '
	call	chrout
	push	hl
	pop	bc
	call	prtdec
	push	hl
	ld	hl,lenmsg
	call	print
	ld	de,infcb
	call	fclose
	pop	hl
rdfilx:	pop	bc
	pop	de
	ret
;
; Write file from buffer.
; On entry, HL points to file name.
; On return, HL = number of bytes written.
;
wrfile:	push	de
	push	bc
	push	hl
	ld	hl,wrmsg
	call	print
	pop	hl
	push	hl
	call	print
	pop	hl
	ld	a,write
	ld	de,outfcb
	call	fopen
	jr	nz,wrfil1
	ld	hl,fopenerrmsg
	call	print
	jr	wrfilx
;
wrfil1:	ld	hl,(bfrptr)
	ld	de,dma
	ld	bc,128
	ldir
	ld	de,outfcb
	ld	c,21
	call	bdos		; write next block
	or	a
	jr	nz,wrfilr
	ld	hl,(bfrptr)
	ld	bc,128
	add	hl,bc
	ld	(bfrptr),hl
	ld	hl,(flen)
	xor	a
	sbc	hl,bc
	ld	(flen),hl
	ld	a,l
	or	h
	jr	nz,wrfil1
	jr	wrfild

wrfilr:	ld	hl,wderrmsg
	call	print
wrfild:	ld	de,outfcb
	call	fclose
wrfilx:	pop	bc
	pop	de
	ret
;
; Output CR/LF to console.
;
crlf:	ld	a,cr
	call	chrout
	ld	a,lf
;
; Output character in A to console.
;
chrout:	push	hl
	push	de
	push	bc
	push	af
	ld	e,a
	ld	c,6
	call	bdos
	pop	af
bdosrt:	pop	bc
	pop	de
	pop	hl
	ret
;
; Wait for input from console, then return character in A.
;
chrin:	call	kbscan
	jr	z,chrin
	ret
;
; Test for console input.
; Return Z if no input, else return NZ.
;
kbscan:	push	hl
	push	de
	push	bc
	ld	e,0ffh
	ld	c,6
	call	bdos
	or	a
	jr	bdosrt
;
; Output null terminated string pointed to by HL.
;
print:	ld	a,(hl)
	inc	hl
	or	a
	ret	z
	call	chrout
	jr	print
;
; Output unsigned number in BC in decimal
; leading zeros are suppressed
;
prtdec:	push	hl
	push	de
	push	bc
	ld	h,b
	ld	l,c
	ld	d,0
	ld	bc,10000
	call	pdecs
	call	pdecc
	ld	bc,1000
	call	pdecs
	call	pdecc
	ld	bc,100
	call	pdecs
	call	pdecc
	ld	bc,10
	call	pdecs
	call	pdecc
	ld	a,l
	add	'0'
	call	chrout
	pop	bc
	pop	de
	pop	hl
	ret
;
pdecc:	or	a
	jr	nz,pdecc1
	or	d
	ret	z
	xor	a
	jr	pdecc2
;
pdecc1:	ld	d,1
pdecc2:	add	'0'
	call	chrout
	ret
;
pdecs:	ld	e,0ffh
pdecs1:	inc	e
	or	a		; clear carry
	sbc	hl,bc
	jr	nc,pdecs1
	add	hl,bc
	ld	a,e
	ret
;
;  Parse command line.
;  Copies command line to CMDLIN and parses it.
;  Sets addresses of arguments in ARGV argument table.
;
parse:	ld	hl,dma		; test for arg input
	ld	de,cmdlin
	ld	b,(hl)		; get command line byte count
	inc	hl
	ld	a,b
	or	a
	jr	nz,parse1	; go if byte count != 0
	ld	e,0		; else argc = 0
	jr	parsex		; and done
;
parse1:	ld	a,(hl)		; transfer command line
	ld	(de),a		; to input buffer
	inc	hl
	inc	de
	djnz	parse1		; loop til done
	xor	a
	ld	(de),a		; terminate input
	ld	hl,cmdlin	; point to input buffer
	ld	e,0		; default to argc = 0
	ld	bc,argv
parse2:	ld	a,(hl)		; skip leading spaces
	inc	hl
	cp	' '
	jr	z,parse2	; loop 'til past spaces
	or	a
	jr	z,parsex	; done if null byte
	dec	hl		; point back to first valid char
	ld	a,l
	ld	(bc),a		; save address of argv[x]
	inc	bc
	ld	a,h
	ld	(bc),a
	inc	bc
	inc	e		; increment argc
parse3:	ld	a,(hl)		; search for end of arg
	inc	hl
	or	a
	jr	z,parsex	; go if end of command line
	cp	' '
	jr	nz,parse3	; loop if not end of string
	dec	hl
	ld	(hl),0		; terminate argv string
	inc	hl
	jr	parse2		; go get next argv
;
parsex:	ld	hl,argc		; save argument count
	ld	(hl),e
	ret
;
noargmsg:
	db	'No command line arguments',CR,LF,0
;
rdmsg:	db	cr,lf,'Reading ',0
wrmsg:	db	cr,lf,'Writing ',0
lenmsg:	db	' bytes read',0
;
rderrmsg:
	db	cr,lf,'ERROR READING FILE!',cr,lf,0
wderrmsg:
	db	cr,lf,'ERROR WRITING FILE!',cr,lf,0
fopenerrmsg:
	db	'  fopen failed',cr,lf,0
;
	defs	100
;
stack	equ	$
;
buffer	equ	$
;
	end
;

